
#include "ptf.h"
#include "ptf_util.h"
#include "ptf_str.h"
#include "ptf_date.h"
#include "ptf_types.h"
#include "ptf_file.h"
#include "ptf_octree.h"
#include "ptf_spheroid.h"
#include "ptf_content.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <vector>
#include <typeinfo>
#include <assert.h>
#include <math.h>

#if defined(_MSC_VER)
#define getcwd _getcwd
#include <direct.h>
#else
#include <unistd.h>
#endif


//  WARNING:
//  This tool is not robust in the face of deliberatly mal-constructed 
//  files, or files with corrupted headers or data. It serves only as 
//  an example of how to operate on a well-formed PTF container file.

//  TODO: Factor reading meta-toc, type toc, and content blocks

char const * indent(size_t level)
{
  if (level > 50) {
    fprintf(stderr, "error: indent level %ld is greater than max supported value 50.\n", level);
    exit(4);
  }
  static char spaces[] = "                                                  ";
  return &spaces[50-level];
}

void check_encoding(ptf_toc_entry const & pte)
{
  if (pte.compression) {
    fprintf(stderr, "compressed files are not supported\n");
    exit(3);
  }
  if (pte.encryption) {
    fprintf(stderr, "encrypted files are not supported\n");
    exit(3);
  }
  if (pte.size_disk != pte.size_mem) {
    fprintf(stderr, "size_disk is different from size_mem: bad file!\n");
    exit(3);
  }
  if (pte.offset == 0) {
    fprintf(stderr, "pte.offset is zero -- unexpected empty chunk\n");
    exit(3);
  }
}

ptf_type cvt_type(char const * type)
{
  ptf_type i_type = 0;
  if (isdigit(*type)) {
    char * tmp;
    if (type[0] == '0' && type[1] == 'x') {
      i_type = strtoul(type+2, &tmp, 16);
    }
    else {
      i_type = strtol(type, &tmp, 16);
    }
    if (!tmp || *tmp != 0) {
      fprintf(stderr, "argument '%s' is not a proper integer\n", type);
      exit(1);
    }
  }
  else {
    fprintf(stderr, "argument '%s' not recognized (should be integer)\n", type);
    exit(1);
  }
  return i_type;
}

template<typename T>
void limit_get(char const * & ptr, char const * end, T & t)
{
  if (end-ptr < sizeof(t)) {
    fprintf(stderr, "limit_get(%s): no data in record\n", typeid(t).name());
    exit(3);
  }
  memcpy(&t, ptr, sizeof(t));
  ptr += sizeof(t);
}


void get_header(char const * fn, FILE * f, ptf_file_header & hdr)
{
  if (impl::fseek64(f, 0, 0)) {
    fprintf(stderr, "%s: seek error\n", fn);
    exit(3);
  }
  if (impl::fread64(&hdr, sizeof(hdr), f) != peSuccess) {
    fprintf(stderr, "%s: read header error\n", fn);
    exit(3);
  }
  if (strncmp((char const *)hdr.magic, PTF_MAGIC_CHARS, sizeof(hdr.magic))) {
    fprintf(stderr, "%s: magic error: not PTF file\n", fn);
    exit(3);
  }
  if (hdr.min_version > PTF_SPECIFICATION_VERSION) {
    fprintf(stderr, "%s: version error: bad PTF file\n", fn);
    exit(3);
  }
  if ((hdr.alignment < 8) || ((hdr.alignment & (hdr.alignment-1)) != 0)) {
    fprintf(stderr, "%s: header alignment error: %d not a valid alignment\n", fn, hdr.alignment);
    exit(3);
  }
}

struct type_record {
  ptf_64 offset;
  ptf_64 size;
  ptf_32 count;
  ptf_toc_entry * toc;
};

//  If the type doesn't exist, create it. Assume that there is enough space 
//  in the meta TOC to append the type.
void get_or_make_type(char const * fn, FILE * f, ptf_file_header & hdr, ptf_meta_toc_entry * mte, ptf_type type, type_record * toc)
{
  toc->offset = 0;
  toc->size = 0;
  toc->count = 0;
  toc->toc = 0;

  for (ptf_32 i = 0; i < hdr.meta_toc.size_mem / sizeof(ptf_meta_toc_entry); ++i) {
    if (mte[i].type == type) {
      //  read this entry
      toc->offset = mte[i].offset;
      toc->count = mte[i].size_mem / sizeof(ptf_toc_entry);
      toc->size = (ptf_64)toc->count * sizeof(ptf_toc_entry);
      //  allow space for adding 4 entries
      ptf_64 size = toc->size + sizeof(ptf_toc_entry) * 4;
      if (sizeof(size_t) < 8 && size > 0x10000000) {
        fprintf(stderr, "%s: file too big for machine\n", fn);
        exit(2);
      }
      toc->toc = (ptf_toc_entry *)malloc((size_t)size);
      if (!toc->toc) {
        fprintf(stderr, "%s: out of memory allocating type toc\n", fn);
        exit(2);
      }
      if (toc->count) {
        if (impl::fseek64(f, toc->offset, 0)) {
          fprintf(stderr, "%s: error seeking to type toc\n", fn);
          exit(2);
        }
        if (impl::fread64(toc->toc, (ptf_64)toc->size, f) != peSuccess) {
          fprintf(stderr, "%s: error reading type toc\n", fn);
          exit(2);
        }
      }
      else {
        toc->offset = 0;
      }
      //  the type is gotten
      return;
    }
  }

  //  gotta add this type
  size_t meta_count = hdr.meta_toc.size_mem / sizeof(ptf_meta_toc_entry);
  memset(&mte[meta_count], 0, sizeof(mte[meta_count]));
  mte[meta_count].type = type;
  hdr.meta_toc.size_disk = hdr.meta_toc.size_mem = (ptf_32)((meta_count + 1) * sizeof(ptf_meta_toc_entry));
  //  make space for at least 4 records
  toc->toc = (ptf_toc_entry *)malloc(sizeof(ptf_toc_entry) * 4);
  if (!toc->toc) {
    fprintf(stderr, "%s: out of memory allocating type toc\n", fn);
    exit(2);
  }
  memset(toc->toc, 0, sizeof(ptf_toc_entry) * 4);
}

void append_toc(type_record * type, ptf_64 offset, ptf_64 size)
{
  if ((ptf_64)(ptf_32)size != size) {
    fprintf(stderr, "Size of record is too big!\n");
    exit(2);
  }
  type->toc[type->count].offset = offset;
  type->toc[type->count].size_mem = type->toc[type->count].size_disk = (ptf_32)size;
  type->count++;
}

void update_type_toc(ptf_meta_toc_entry * mte, ptf_type type, type_record * rec)
{
  for (ptf_32 i = 0; mte->type != type; ++i) {
    ++mte;
  }
  mte->size_mem = mte->size_disk = rec->count * sizeof(ptf_toc_entry);
  mte->offset = rec->offset;
}

void pad_file(char const * fn, FILE * f, ptf_file_header & hdr)
{
  if (impl::fseek64(f, 0, 2)) {
    fprintf(stderr, "%s: cannot seek to end\n", fn);
    exit(2);
  }
  ptf_64 pos = impl::ftell64(f);
  if (pos >= hdr.size) {
    //  already have the length I need
    return;
  }
  if (pos < hdr.size-hdr.alignment) {
    //  file is corrupt
    fprintf(stderr, "%s: file is too short\n", fn);
    exit(3);
  }
  void * data = calloc(hdr.alignment, 1);
  if (impl::fwrite64(data, hdr.size-pos, f) != peSuccess) {
    fprintf(stderr, "%s: error padding at end\n", fn);
    exit(3);
  }
  free(data);
}



void dump_toc(char const * fn, ptf_file_header const & hdr, ptf_meta_toc_entry & mte, FILE * f)
{
  fprintf(stdout, "\n");
  fprintf(stdout, "TOC type 0x%08x (%d entries):\n", mte.type, mte.size_mem/sizeof(ptf_meta_toc_entry));
  fprintf(stdout, "%ld bytes on disk, %ld bytes in memory\n", mte.size_disk, mte.size_mem);
  if( mte.compression ) {
    fprintf(stdout, "Uses compression type %d, ", mte.compression );
  }
  else {
    fprintf(stdout, "Not compressed, " );
  }
  if( mte.encryption ) {
    fprintf(stdout, "uses encryption type %d, ", mte.encryption );
  }
  else {
    fprintf(stdout, "not encrypted, " );
  }
  if( mte.hash ) {
    fprintf(stdout, "Checksum 0x%llx\n", mte.hash );
  }
  else {
    fprintf(stdout, "Not checksummed\n" );
  }
  if (mte.offset + mte.size_disk > hdr.size) {
    fprintf(stderr, "%s: TOC offset 0x%llx+0x%x*%d (or header size 0x%llx) is wrong\n",
        fn, mte.offset, mte.size_mem/sizeof(ptf_meta_toc_entry), sizeof(ptf_toc_entry), hdr.size);
    exit(4);
  }
  std::vector<ptf_toc_entry> toc;
  toc.resize(mte.size_mem/sizeof(ptf_meta_toc_entry));
  if (mte.size_mem/sizeof(ptf_meta_toc_entry)) {
    if (impl::fseek64(f, mte.offset, 0)) {
      fprintf(stderr, "%s: seek error\n", fn);
      exit(3);
    }
    if (impl::fread64(&toc[0], toc.size() * (ptf_64)sizeof(ptf_toc_entry), f) != peSuccess) {
      fprintf(stderr, "%s: read error\n", fn);
      exit(3);
    }
    for (size_t n = toc.size(), i = 0; i != n; ++i) {
      fprintf(stdout, "[%d] record size 0x%lx offset 0x%llx\n", i, (size_t)toc[i].size_mem, toc[i].offset);
      if (toc[i].size_disk + toc[i].offset > hdr.size) {
        fprintf(stderr, "  record out of range! (size_disk 0x%llx)\n", toc[i].size_disk);
        exit(3);
      }
      if (toc[i].size_disk && !toc[i].offset) {
        fprintf(stderr, "  record without offset!\n");
        exit(3);
      }
      if (toc[i].size_disk > toc[i].size_mem) {
        fprintf(stderr, "  record is bigger on disk (0x%llx) than in memory (0x%llx)\n",
            (ptf_64)toc[i].size_disk, (ptf_64)toc[i].size_mem);
        exit(3);
      }
      if (!toc[i].compression && (toc[i].size_mem != toc[i].size_disk)) {
        fprintf(stderr, "  record size mis-match between disk (0x%llx) and memory (0x%llx)\n",
            (ptf_64)toc[i].size_disk, (ptf_64)toc[i].size_mem);
        exit(3);
      }
      if( toc[i].compression ) {
        fprintf(stdout, "Uses compression type %d, ", toc[i].compression );
      }
      else {
        fprintf(stdout, "Not compressed, " );
      }
      if( toc[i].encryption ) {
        fprintf(stdout, "uses encryption type %d, ", toc[i].encryption );
      }
      else {
        fprintf(stdout, "not encrypted, " );
      }
      if( toc[i].hash ) {
        fprintf(stdout, "Checksum 0x%llx\n", toc[i].hash );
      }
      else {
        fprintf(stdout, "Not checksummed\n" );
      }
    }
  }
}

void dump_meta(char const * fn, ptf_file_header const & hdr, FILE * f)
{
  fprintf(stdout, "\n");
  fprintf(stdout, "Meta-TOC (%d entries):\n", hdr.meta_toc.size_mem/sizeof(ptf_meta_toc_entry));
  if (hdr.meta_toc.offset + hdr.meta_toc.size_disk > hdr.size) {
    fprintf(stderr, "%s: file size %llu (or meta-toc (%llu+%llu)) is wrong\n",
        fn, hdr.size, hdr.meta_toc.offset, hdr.meta_toc.size_disk);
    exit(4);
  }
  if (hdr.meta_toc.size_mem == 0) {
    return;
  }
  std::vector<ptf_meta_toc_entry> meta;
  meta.resize(hdr.meta_toc.size_mem/sizeof(ptf_meta_toc_entry));
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: seek error\n", fn);
    exit(3);
  }
  if (impl::fread64(&meta[0], hdr.meta_toc.size_mem, f) != peSuccess) {
    fprintf(stderr, "%s: read error\n", fn);
    exit(3);
  }
  for (size_t i = 0; i < meta.size(); ++i) {
    fprintf(stdout, "Type 0x%08x, %d entries offset 0x%llx\n",
        meta[i].type, meta[i].size_mem/sizeof(ptf_toc_entry), meta[i].offset);
  }
  if( hdr.meta_toc.compression ) {
    fprintf(stdout, "Uses compression type %d, ", hdr.meta_toc.compression );
  }
  else {
    fprintf(stdout, "Not compressed, " );
  }
  if( hdr.meta_toc.encryption ) {
    fprintf(stdout, "uses encryption type %d, ", hdr.meta_toc.encryption );
  }
  else {
    fprintf(stdout, "not encrypted, " );
  }
  if( hdr.meta_toc.hash ) {
    fprintf(stdout, "Checksum 0x%llx\n", hdr.meta_toc.hash );
  }
  else {
    fprintf(stdout, "Not checksummed\n" );
  }
  for (size_t i = 0; i < meta.size(); ++i) {
    dump_toc(fn, hdr, meta[i], f);
  }
}


ptf_file_header hdr;

void dump_header(char const * fn, FILE * f)
{
  fprintf(stdout, "\n");
  get_header(fn, f, hdr);
  fprintf(stdout, "%s: PTF version %x.%x.%x readable by %x.%x.%x\n", fn, 
      hdr.spec_version>>16, (hdr.spec_version>>8)&0xff, hdr.spec_version&0xff, 
      hdr.min_version>>16, (hdr.min_version>>8)&0xff, hdr.min_version&0xff);
  char fname[256];
  ptf_str_to_char(hdr.file_name, fname, 256, '?');
  fprintf(stdout, "canonical name    : %s\n", fname);
  fprintf(stdout, "creation date     : %04d-%02d-%02d %02d:%02d:%02d.%02d\n",
      hdr.cr_date.year, hdr.cr_date.month, hdr.cr_date.day, hdr.cr_date.hour, hdr.cr_date.minute,
      hdr.cr_date.second, hdr.cr_date.hundredth);
  fprintf(stdout, "modification date : %04d-%02d-%02d %02d:%02d:%02d.%02d\n",
      hdr.mod_date.year, hdr.mod_date.month, hdr.mod_date.day, hdr.mod_date.hour, hdr.mod_date.minute,
      hdr.mod_date.second, hdr.mod_date.hundredth);
  fprintf(stdout, "alignment         : 0x%x\n", hdr.alignment);
  fprintf(stdout, "meta table size   : %d types\n", hdr.meta_toc.size_mem/sizeof(ptf_toc_entry));
  fprintf(stdout, "meta table offset : 0x%llx\n", hdr.meta_toc);
  fprintf(stdout, "file size         : 0x%llx\n", hdr.size);
  fprintf(stdout, "extents           : 0x%x,0x%x,0x%x - 0x%x,0x%x,0x%x\n",
      hdr.range.x_min, hdr.range.y_min, hdr.range.z_min, 
      hdr.range.x_max, hdr.range.y_max, hdr.range.z_max);
}


void dump_octree_block(char const * path, FILE * f, ptf_toc_entry const * pte, size_t n_toc, size_t index, size_t depth, size_t left, size_t bottom, size_t yon, size_t side);

void dump_octree_node(char const * path, FILE * f, ptf_toc_entry const * pte, size_t n_toc, size_t index, size_t node, size_t depth, char const * block, size_t blockSize, size_t left, size_t bottom, size_t yon, size_t side)
{
  ptf_octree_block_header const * obh = (ptf_octree_block_header const *)block;

  if (obh->level > 31) {
    fprintf(stderr, "%s: corrupted block (level out of bounds)\n", path);
    exit(3);
  }
  if (obh->content_array_pointer + obh->num_nodes * sizeof(ptf_fixed_ref) > blockSize) {
    fprintf(stderr, "%s: corrupted block (content_array_pointer)\n", path);
    exit(3);
  }
  if (obh->link_pointer + obh->num_links * sizeof(ptf_remote_ref) > blockSize) {
    fprintf(stderr, "%s: corrupted block (num_links)\n", path);
    exit(3);
  }
  if (obh->node_array_pointer + obh->num_nodes * sizeof(ptf_octree_node) > blockSize) {
    fprintf(stderr, "%s: corrupted block (num_nodes)\n", path);
    exit(3);
  }

  ptf_octree_node const * on = (ptf_octree_node const *)(block + obh->node_array_pointer);
  ptf_fixed_ref const * fr = (ptf_fixed_ref const *)(block + obh->content_array_pointer);
  ptf_remote_ref const * rr = (ptf_remote_ref const *)(block + obh->link_pointer);

  fprintf(stdout, "%s%ld:%ld[", indent(depth), index, node);
  short const * sc = (short const *)on[node].child;
  for (int i = 0; i < 8; ++i) {
    if (sc[i] & 0x8000) {
      fprintf(stdout, "%ld:%ld:0", rr[sc[i]&0x7fff].file, rr[sc[i]&0x7fff].index);
    }
    else if (sc[i] > 0) {
      fprintf(stdout, "0:%ld", sc[i]);
    }
    else {
      fprintf(stdout, "-");
    }
    if (i < 7) {
      fprintf(stdout, ",");
    }
  }
  fprintf(stdout, "],  {0x%08x,0x%08x,0x%08x}-{0x%08x,0x%08x,0x%08x}\n",
      left, bottom, yon, left+side, bottom+side, yon+side);
  for (int i = 0; i < 8; ++i) {
    int dx = (i & 4) ? 1 : 0;
    int dy = (i & 2) ? 1 : 0;
    int dz = (i & 1) ? 1 : 0;
    if (sc[i] & 0x8000) {
      if (rr[sc[i]&0x7fff].file != 0) {
        //  that's just a warning -- keep going, after printing a message
        fprintf(stdout, "warning: remote reference %ld:%ld not followed\n", 
            rr[sc[i]&0x7fff].file, rr[sc[i]&0x7fff].index);
      }
      else {
        dump_octree_block(path, f, pte, n_toc, rr[sc[i]&0x7fff].index, depth+1, 
            left + dx*(side>>1), bottom + dy*(side>>1), yon + dz*(side>>1), side>>1);
      }
    }
    else if (sc[i] > 0) {
      dump_octree_node(path, f, pte, n_toc, index, sc[i], depth+1, block, blockSize, 
            left + dx*(side>>1), bottom + dy*(side>>1), yon + dz*(side>>1), side>>1);
    }
  }
}

void dump_octree_block(char const * path, FILE * f, ptf_toc_entry const * pte, size_t n_toc, size_t index, size_t depth,
  size_t left, size_t bottom, size_t yon, size_t side)
{
  char * block = 0;
  if (index >= n_toc) {
    fprintf(stderr, "%s: octree TOC index %ld out of range (%ld)\n", path, index, n_toc);
    exit(3);
  }
  if (pte[index].compression || pte[index].encryption) {
    fprintf(stderr, "%s: ptfdump does not support encrypted/compressed records\n", path);
    exit(3);
  }
  if (impl::fseek64(f, pte[index].offset, 0)) {
    fprintf(stderr, "%s: seek(octree_toc.offset: %llu) failed\n", path, pte[index].offset);
    exit(3);
  }
  block = new char[pte[index].size_mem];
  if (impl::fread64(block, pte[index].size_mem, f) != peSuccess) {
    fprintf(stderr, "%s: short read on octree block\n", path);
    exit(3);
  }
  dump_octree_node(path, f, pte, n_toc, index, 0, depth, block, pte[index].size_mem, left, bottom, yon, side);
  delete[] block;
}

void dump_tree(char const * path, ptf_file_header const & hdr, FILE * f)
{
  check_encoding(hdr.meta_toc);
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: seek(meta_toc.offset %llu) failed\n", path, hdr.meta_toc.offset);
    exit(3);
  }
  size_t n_meta = hdr.meta_toc.size_mem / sizeof(ptf_meta_toc_entry);
  ptf_meta_toc_entry * mte = new ptf_meta_toc_entry[n_meta];
  if (impl::fread64(mte, sizeof(ptf_meta_toc_entry) * n_meta, f) != peSuccess) {
    fprintf(stderr, "%s: short read of meta TOC\n", path);
    exit(3);
  }
  size_t i;
  for (i = 0; i < n_meta; ++i) {
    if (mte[i].type == ptOctreeNode) {
      goto found_it;
    }
  }
  fprintf(stderr, "%s: no octree found\n", path);
  exit(3);
found_it:
  check_encoding(mte[i]);
  if (impl::fseek64(f, mte[i].offset, 0)) {
    fprintf(stderr, "%s: seek(octree.offset: %llu) failed\n", path, mte[i].offset);
    exit(3);
  }
  size_t n_toc = mte[i].size_mem / sizeof(ptf_toc_entry);
  ptf_toc_entry * pte = new ptf_toc_entry[n_toc];
  if (impl::fread64(pte, n_toc * sizeof(ptf_toc_entry), f) != peSuccess) {
    fprintf(stderr, "%s: short read of octree TOC\n", path);
    exit(3);
  }
  dump_octree_block(path, f, pte, n_toc, 1, 0, hdr.range.x_min, hdr.range.y_min, hdr.range.z_min, hdr.range.x_max - hdr.range.x_min);
  delete[] mte;
  delete[] pte;
}

void add_block(char const * fn, FILE * f, char const * type, char const * blockname)
{
  //  convert argument
  ptf_type i_type = cvt_type(type);

  //  open up the file
  ptf_file_header hdr;
  get_header(fn, f, hdr);

  //  read the input data
  FILE * f2 = fopen(blockname, "rb");
  if (!f2) {
    fprintf(stderr, "%s: file not found\n", blockname);
    exit(2);
  }
  impl::fseek64(f2, 0, 2);
  ptf_64 size = impl::ftell64(f2);
  if (sizeof(size_t) < 8 && size > 0x10000000) {
    fprintf(stderr, "%s: file too big for machine\n", blockname);
    exit(2);
  }
  void * data = malloc((size_t)size);
  if (data == 0) {
    fprintf(stderr, "%s: out of memory (%lld bytes)\n", blockname, size);
    exit(2);
  }
  impl::fseek64(f2, 0, 0);
  if (impl::fread64(data, size, f2) != peSuccess) {
    fprintf(stderr, "%s: error reading file\n", blockname);
    exit(2);
  }
  fclose(f2);

  //  read the meta-toc
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: can't seek to meta-toc\n", fn);
    exit(3);
  }
  check_encoding(hdr.meta_toc);
  //  allow expansion by two types (the type being added, and the free chunk table, 
  //  if not already present).
  ptf_64 to_read = (ptf_64)hdr.meta_toc.size_mem;
  if (sizeof(size_t) == 4 && to_read > 0x10000000) {
    fprintf(stderr, "%s: file too big for machine\n", fn);
    exit(3);
  }
  ptf_meta_toc_entry * mte = (ptf_meta_toc_entry *)malloc((size_t)to_read 
      + 2 * sizeof(ptf_meta_toc_entry));
  if (!mte) {
    fprintf(stderr, "%s: out of memory for meta toc\n", fn);
    exit(2);
  }
  if (impl::fread64(mte, to_read, f) != peSuccess) {
    fprintf(stderr, "%s: error reading meta toc\n", fn);
    exit(2);
  }

  //  read the free-data TOC, if present
  type_record empty_records;
  get_or_make_type(fn, f, hdr, mte, ptEmptyRecord, &empty_records);

  //  read the target type TOC, if present
  type_record type_records;
  get_or_make_type(fn, f, hdr, mte, i_type, &type_records);

  //  add chunk at end of file
  pad_file(fn, f, hdr);
  if (impl::fseek64(f, hdr.size, 0)) {
    fprintf(stderr, "%s: can't seek to end\n", fn);
    exit(2);
  }
  if (impl::fwrite64(data, size, f) != peSuccess) {
    fprintf(stderr, "%s: can't append data\n", fn);
    exit(2);
  }
  type_records.toc[type_records.count].offset = hdr.size;
  type_records.toc[type_records.count].size_mem = type_records.toc[type_records.count].size_disk = (ptf_32)size;
  type_records.count++;
  hdr.size = (hdr.size + size + hdr.alignment-1) & ~((ptf_64)hdr.alignment-1);

  //  retire current free chunk, type toc, and meta toc
  if (empty_records.size) {
    append_toc(&empty_records, empty_records.offset, empty_records.size);
  }
  if (type_records.size) {
    append_toc(&empty_records, type_records.offset, type_records.size);
  }
  if (hdr.meta_toc.offset) {
    append_toc(&empty_records, hdr.meta_toc.offset, (ptf_64)hdr.meta_toc.size_mem);
  }

  //  write new type toc
  pad_file(fn, f, hdr);
  type_records.offset = hdr.size;
  type_records.size = (ptf_64)type_records.count * sizeof(ptf_toc_entry);
  if (impl::fseek64(f, type_records.offset, 0)) {
    fprintf(stderr, "%s: can't seek to end\n", fn);
    exit(2);
  }
  if (impl::fwrite64(type_records.toc, type_records.size, f) != peSuccess) {
    fprintf(stderr, "%s: can't append type toc\n", fn);
    exit(2);
  }
  hdr.size = (type_records.offset + type_records.size + (ptf_64)hdr.alignment - 1)
      & ~((ptf_64)hdr.alignment - 1);
  update_type_toc(mte, i_type, &type_records);

  //  write new free toc
  pad_file(fn, f, hdr);
  empty_records.offset = hdr.size;
  empty_records.size = (ptf_64)empty_records.count * sizeof(ptf_toc_entry);
  if (impl::fseek64(f, empty_records.offset, 0)) {
    fprintf(stderr, "%s: can't seek to end\n", fn);
    exit(2);
  }
  if (impl::fwrite64(empty_records.toc, empty_records.size, f) != peSuccess) {
    fprintf(stderr, "%s: can't append empty record toc\n", fn);
    exit(2);
  }
  hdr.size = (empty_records.offset + empty_records.size + (ptf_64)hdr.alignment - 1)
      & ~((ptf_64)hdr.alignment - 1);
  update_type_toc(mte, ptEmptyRecord, &empty_records);

  //  write new meta toc
  pad_file(fn, f, hdr);
  hdr.meta_toc.offset = hdr.size;
  size = (ptf_64)hdr.meta_toc.size_disk;
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: can't seek to meta-toc at end\n", fn);
    exit(2);
  }
  if (impl::fwrite64(mte, size, f) != peSuccess) {
    fprintf(stderr, "%s: can't append meta toc\n", fn);
    exit(2);
  }
  hdr.size = (hdr.meta_toc.offset + size + (ptf_64)hdr.alignment - 1)
      & ~((ptf_64)hdr.alignment - 1);

  //  update header
  if (impl::fseek64(f, 0, 0)) {
    fprintf(stderr, "%s: can't seek to header\n", fn);
    exit(2);
  }
  if (impl::fwrite64(&hdr, sizeof(hdr), f) != peSuccess) {
    fprintf(stderr, "%s: can't write header\n", fn);
    exit(2);
  }
}

void remove_block(char const * fn, FILE * f, char const * type, char const * index)
{
  //  convert argument
  ptf_type i_type = cvt_type(type);
  ptf_32 i_index = cvt_type(index);

  //  open up the file
  ptf_file_header hdr;
  get_header(fn, f, hdr);


}

void dump_one_mesh(char const * fn, FILE * f, ptf_32 i, ptf_toc_entry const & pte)
{
  check_encoding(pte);
  if (impl::fseek64(f, pte.offset, 0)) {
    fprintf(stderr, "%s: error seeking to mesh %d\n", fn, i);
    exit(3);
  }
  char * data = (char *)malloc(pte.size_mem);
  if (!data) {
    fprintf(stderr, "%s: mesh %d out of memory\n", fn, i);
    exit(3);
  }
  if (impl::fread64(data, pte.size_mem, f) != peSuccess) {
    fprintf(stderr, "%s: error reading mesh %d\n", fn, i);
    exit(3);
  }
  char const * ptr = data;
  char const * end = ptr + pte.size_mem;
  ptf_mesh_header header;
  // limit_get will read from ptr into header the sizeof(header)
  // it will also mover ptr to ptr+sizeof(header)
  limit_get(ptr, end, header);
  fprintf(stderr, "Mesh %d (\"%s\"): %ld vertices, %d components; %d subsets; info at 0x%x\n", i, header.name, header.vertex_count, header.num_components, header.num_subsets, header.data_offset);

  ptf_32 positionComponentOffset = 0;
  ptf_32 positionComponentCount = 0;
  ptf_32 normalComponentOffset = 0;
  ptf_32 normalComponentCount = 0;
  ptf_32 colorComponentOffset = 0;
  ptf_32 colorComponentCount = 0;
  ptf_32 componentOffset = 0;
  for (ptf_32 j = 0; j < header.num_components; ++j) {
    ptf_vertex_component_description component;
    limit_get(ptr, end, component);
    fprintf(stderr, "  Component %d: Usage 0x%x index %d; Type 0x%x count %d\n", 
        j, component.usage, component.usage_index, component.type, component.type_count);
    if( component.usage == pvuPosition ) {
      positionComponentOffset = componentOffset;
      positionComponentCount = component.type_count;
      if( component.type != pvtFloat32 ) {
        fprintf(stderr, "  Non-float position types currently not supported: Component %d: Usage 0x%x index %d; Type 0x%x count %d\n", 
            j, component.usage, component.usage_index, component.type, component.type_count);
        exit(3);
      }
	} 
    else if (component.usage == pvuColor) {
		colorComponentOffset = componentOffset;
		colorComponentCount = component.type_count;
		if( component.type != pvtUInt8 ) {
			fprintf(stderr, "  color types other than unsigned byte currently not supported: Component %d: Usage 0x%x index %d; Type 0x%x count %d\n", 
				j, component.usage, component.usage_index, component.type, component.type_count);
			exit(3);
		}
	}
    else if (component.usage == pvuNormal) {
		normalComponentOffset = componentOffset;
		normalComponentCount = component.type_count;
        if( component.type != pvtFloat32 ) {
            fprintf(stderr, "  Non-float normal types currently not supported: Component %d: Usage 0x%x index %d; Type 0x%x count %d\n", 
                j, component.usage, component.usage_index, component.type, component.type_count);
            exit(3);
        }
	}
    switch( component.type ) {
      case pvtFloat32: componentOffset += sizeof( float )*component.type_count; break;
      case pvtUInt32: componentOffset += sizeof( ptf_32 )*component.type_count; break;
      case pvtInt32: componentOffset += sizeof( ptf_32 )*component.type_count; break;
      case pvtUInt16: componentOffset += sizeof( ptf_16 )*component.type_count; break;
      case pvtInt16: componentOffset += sizeof( ptf_16 )*component.type_count; break;
      case pvtUInt8: componentOffset += sizeof( ptf_8 )*component.type_count; break;
      case pvtInt8: componentOffset += sizeof( ptf_8 )*component.type_count; break;
    }
  }

  // Readjust the alignment to the file alignment
  // e.g., if the previous limit_get read was of a size of 3, and our alignment is 4
  // then we need to fix it up
  ptr += (ptf_32)(header.file_alignment - 1) & -header.file_alignment;

  ptf_mesh_subset_description * subsets = (ptf_mesh_subset_description *)malloc(header.num_subsets * sizeof(ptf_mesh_subset_description));
  for (ptf_32 j = 0; j < header.num_subsets; ++j) {
    limit_get(ptr, end, subsets[j]);
  }

  char * vertices = (char *)malloc(header.vertex_stride * header.vertex_count);
  memcpy(vertices, ptr, header.vertex_stride * header.vertex_count);
  ptr += header.vertex_stride * header.vertex_count;

  // Verts
  fprintf(stderr, "    Vertices (%ld):\n", header.vertex_count);
  for (ptf_32 i = 0; i < header.vertex_count; ++i) {
    char * val = &vertices[ i * header.vertex_stride ] + positionComponentOffset;
    fprintf(stderr, "      ");
    for (ptf_32 v = 0; v < positionComponentCount; ++v) {
      fprintf(stderr, " %f", *((float*)(val + v*sizeof(float))));
      if( v < (positionComponentCount-1) ) {
        fprintf(stderr, ", ");
      }
    }
	if (normalComponentCount > 0) {
	    char * val2 = &vertices[ i * header.vertex_stride ] + normalComponentOffset;
		fprintf(stderr, " [");
		for (ptf_32 v = 0; v < normalComponentCount; ++v) {
			fprintf(stderr, " %f", *((float*)(val2 + v*sizeof(float))));
			if( v < (normalComponentCount-1) ) {
				fprintf(stderr, ", ");
			}
		}
		fprintf(stderr, " ]");
	}
	if (colorComponentCount > 0) {
	    char * val2 = &vertices[ i * header.vertex_stride ] + colorComponentOffset;
		fprintf(stderr, " [");
		for (ptf_32 v = 0; v < colorComponentCount; ++v) {
			fprintf(stderr, " %d", *((unsigned char*)(val2 + v*sizeof(char))));
			if( v < (colorComponentCount-1) ) {
				fprintf(stderr, ", ");
			}
		}
		fprintf(stderr, " ]");
	}
    fprintf(stderr, "\n");
  }


  // Read and output subset data
  for (ptf_32 j = 0; j < header.num_subsets; ++j) {
    fprintf(stderr, "  Subset %d: Type %d\n", 
        j, subsets[j].primitive_type);
    size_t iSize = 0;
    switch (subsets[j].index_type) {
      case pvtUInt32:
        iSize = 4;
        break;
      case pvtUInt16:
        iSize = 2;
        break;
      default:
        fprintf(stderr, "  Bad index_type in subset %d: index_type %d\n", 
            j, subsets[j].index_type);
        exit(3);
        break;
    }

    // Indices
    char * indices = (char *)malloc(iSize * subsets[j].num_indices);
    memcpy(indices, ptr, iSize * subsets[j].num_indices);
    ptr += (ptf_32)(iSize * subsets[j].num_indices + header.file_alignment - 1)
            & -header.file_alignment;
    ptf_32 * index32 = (ptf_32*)indices;
    ptf_16 * index16 = (ptf_16*)indices;

    fprintf(stderr, "    Indices (%ld):\n", subsets[j].num_indices);
    for (ptf_32 i = 0; i < subsets[j].num_indices;) {
      fprintf(stderr, "     ");
      for (ptf_32 v = 0; v < subsets[j].primitive_type; ++v, ++i) {
        ptf_32 index;
        switch (subsets[j].index_type) {
          case pvtUInt32: index = index32[ i ]; break;
          case pvtUInt16: index = index16[ i ]; break;
        }
        fprintf(stderr, " %ld", index);
      }
      fprintf(stderr, "\n");
    }

    // Use this code to print out the actual vertices
    /*
    char * vertPtr = (vertices + positionComponentOffset);
    for (ptf_32 i = 0; i < subsets[j].num_indices;) {
      fprintf(stderr, "  N-gon %d vertices: \n", 
          i/subsets[j].primitive_type);
      for (ptf_32 v = 0; v < subsets[j].primitive_type; ++v, ++i) {
        ptf_32 index;
        switch (subsets[j].index_type) {
          case pvtUInt32: index = index32[ i ]; break;
          case pvtUInt16: index = index16[ i ]; break;
        }
        float * vertex = (float*)(&vertPtr[ index * header.vertex_stride ]);
        fprintf(stderr, "  %f", *vertex);
      }
      fprintf(stderr, "\n");
    }
    */

    free(indices);
  }

  free(vertices);
  free(subsets);
  free(data);
}

void dump_some_mesh(char const * fn, FILE * f, ptf_32 from, ptf_32 to)
{
  //  open up the file
  ptf_file_header hdr;
  get_header(fn, f, hdr);

  //  read the meta-toc
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: can't seek to meta-toc\n", fn);
    exit(3);
  }
  check_encoding(hdr.meta_toc);
  //  allow expansion by two types (the type being added, and the free chunk table, 
  //  if not already present).
  ptf_64 to_read = (ptf_64)hdr.meta_toc.size_mem;
  if (sizeof(size_t) == 4 && to_read > 0x10000000) {
    fprintf(stderr, "%s: file too big for machine\n", fn);
    exit(3);
  }
  ptf_meta_toc_entry * mte = (ptf_meta_toc_entry *)malloc((size_t)to_read 
      + 2 * sizeof(ptf_meta_toc_entry));
  if (!mte) {
    fprintf(stderr, "%s: out of memory for meta toc\n", fn);
    exit(2);
  }
  if (impl::fread64(mte, to_read, f) != peSuccess) {
    fprintf(stderr, "%s: error reading meta toc\n", fn);
    exit(2);
  }

  //  find the type
  ptf_32 nMeta = hdr.meta_toc.size_mem / sizeof(ptf_meta_toc_entry);
  for (ptf_32 i = 0; i < nMeta; ++i) {
    if (mte[i].type == ptMeshData) {
      // read the TOC
      //  I'm taking advantage of knowing that meta_toc_entry and toc_entry have the same layout
      check_encoding((ptf_toc_entry const &)mte[i]);
      ptf_32 n = mte[i].size_mem / sizeof(ptf_toc_entry);
      if (to >= n) {
        to = n-1;
      }
      if (to < from) {
        fprintf(stderr, "%s: %d is greater than %d (bad arguments)\n", fn, from, to);
        exit(2);
      }
      if (impl::fseek64(f, mte[i].offset, 0)) {
        fprintf(stderr, "%s: bad seek to mesh toc\n", fn);
        exit(3);
      }
      ptf_toc_entry * pte = (ptf_toc_entry *)malloc(n * sizeof(ptf_toc_entry));
      if (impl::fread64(pte, n * sizeof(ptf_toc_entry), f) != peSuccess) {
        fprintf(stderr, "%s: error reading mesh toc\n", fn);
        exit(3);
      }
      do {
        dump_one_mesh(fn, f, from, pte[from]);
      }
      while (from++ != to);
      return;
    }
  }
  fprintf(stderr, "%s: no mesh found\n", fn);
}

void dump_mesh(char const * fn, FILE * f, ptf_32 index)
{
  dump_some_mesh(fn, f, index, index);
}

void dump_all_mesh(char const * fn, FILE * f)
{
  dump_some_mesh(fn, f, 0, 0xffffffff);
}

void dump_one_content(char const * fn, FILE * f, ptf_32 i, ptf_toc_entry const & pte)
{
  check_encoding(pte);
  if (impl::fseek64(f, pte.offset, 0)) {
    fprintf(stderr, "%s: error seeking to content %d\n", fn, i);
    exit(3);
  }
  char * data = (char *)malloc(pte.size_mem);
  if (!data) {
    fprintf(stderr, "%s: content %d out of memory\n", fn, i);
    exit(3);
  }
  if (impl::fread64(data, pte.size_mem, f) != peSuccess) {
    fprintf(stderr, "%s: error reading content %d\n", fn, i);
    exit(3);
  }
  char const * ptr = data;
  char const * end = ptr + pte.size_mem;
  ptf_node_content_header pnch;
  limit_get(ptr, end, pnch);
  fprintf(stderr, "Content %d: %d types; info at 0x%x\n", i, pnch.num_types, pnch.content_info_offset);
  for (ptf_32 j = 0; j < pnch.num_types; ++j) {
    ptf_node_content_type pnct;
    limit_get(ptr, end, pnct);
    fprintf(stderr, "  Type 0x%x Index %d; %d instances; info at 0x%x\n", 
        pnct.type.type, pnct.type.index, pnct.num_instances, pnct.type_info_offset);
    for (ptf_32 q = 0; q < pnct.num_instances; ++q) {
      ptf_node_content_instance pnci;
      limit_get(ptr, end, pnci);
      fprintf(stderr, "    Instance %d:%d: pos(%.2f,%.2f,%.2f) ori({%.3f,%.3f,%.3f},%.3f)\n",
          j, q, pnci.pos.x, pnci.pos.y, pnci.pos.z, pnci.ori.x, pnci.ori.y, pnci.ori.z, pnci.ori.w);
    }
  }
  free(data);
}

void dump_some_content(char const * fn, FILE * f, ptf_32 from, ptf_32 to)
{
  //  open up the file
  ptf_file_header hdr;
  get_header(fn, f, hdr);

  //  read the meta-toc
  if (impl::fseek64(f, hdr.meta_toc.offset, 0)) {
    fprintf(stderr, "%s: can't seek to meta-toc\n", fn);
    exit(3);
  }
  check_encoding(hdr.meta_toc);
  //  allow expansion by two types (the type being added, and the free chunk table, 
  //  if not already present).
  ptf_64 to_read = (ptf_64)hdr.meta_toc.size_mem;
  if (sizeof(size_t) == 4 && to_read > 0x10000000) {
    fprintf(stderr, "%s: file too big for machine\n", fn);
    exit(3);
  }
  ptf_meta_toc_entry * mte = (ptf_meta_toc_entry *)malloc((size_t)to_read 
      + 2 * sizeof(ptf_meta_toc_entry));
  if (!mte) {
    fprintf(stderr, "%s: out of memory for meta toc\n", fn);
    exit(2);
  }
  if (impl::fread64(mte, to_read, f) != peSuccess) {
    fprintf(stderr, "%s: error reading meta toc\n", fn);
    exit(2);
  }

  //  find the type
  ptf_32 nMeta = hdr.meta_toc.size_mem / sizeof(ptf_meta_toc_entry);
  for (ptf_32 i = 0; i < nMeta; ++i) {
    if (mte[i].type == ptOctreeContent) {
      // read the TOC
      //  I'm taking advantage of knowing that meta_toc_entry and toc_entry have the same layout
      check_encoding((ptf_toc_entry const &)mte[i]);
      ptf_32 n = mte[i].size_mem / sizeof(ptf_toc_entry);
      if (to >= n) {
        to = n-1;
      }
      if (to < from) {
        fprintf(stderr, "%s: %d is greater than %d (bad arguments)\n", fn, from, to);
        exit(2);
      }
      if (impl::fseek64(f, mte[i].offset, 0)) {
        fprintf(stderr, "%s: bad seek to content toc\n", fn);
        exit(3);
      }
      ptf_toc_entry * pte = (ptf_toc_entry *)malloc(n * sizeof(ptf_toc_entry));
      if (impl::fread64(pte, n * sizeof(ptf_toc_entry), f) != peSuccess) {
        fprintf(stderr, "%s: error reading content toc\n", fn);
        exit(3);
      }
      do {
        dump_one_content(fn, f, from, pte[from]);
      }
      while (from++ != to);
      return;
    }
  }
  fprintf(stderr, "%s: no octree content found\n", fn);
}

void dump_content(char const * fn, FILE * f, ptf_32 index)
{
  dump_some_content(fn, f, index, index);
}

void dump_all_content(char const * fn, FILE * f)
{
  //  Don't dump content 0, because that's "special" and won't exist.
  dump_some_content(fn, f, 1, 0xffffffff);
}

/****************************************************************
 * These functions are required for insertspheroid command line *
 ****************************************************************/

bool cubeTooBig( float span, ptf_ocube const & cube )
{
  float cubeSpan;
  
  ptf_geopos cubeMax, cubeMin;
  ptf_err err = ptf_octree_box_get_coords( cube, &cubeMax, &cubeMin );
  cubeSpan = fabs( cubeMax.x - cubeMin.x );

  return cubeSpan > ( span * 2.0 );
}

bool octreeBlockTooBig( ptf_ocube const & box, ptf_ocube const & cube )
{
  ptf_geopos boxMax, boxMin;
  ptf_err err = ptf_octree_box_get_coords( box, &boxMax, &boxMin );
  float boxSpan = fabs( boxMax.x - boxMin.x );
  return cubeTooBig( boxSpan, cube );
}

ptf_err splitOctreeToFitCube( ptf_octree_handle const & octree, ptf_onode_ref const & node, ptf_ocube const & cube, ptf_onode_ref * o_node )
{
  bool blockTooBig = octreeBlockTooBig( cube, node.extent );
  if( blockTooBig ) {
    ptf_geopos lower, upper;
    ptf_err err = ptf_octree_box_get_coords( cube, &lower, &upper );
    if( err != peSuccess ) {
      return err;
    }

    ptf_geopos center;
    center.x = ( lower.x + upper.x ) / 2;
    center.y = ( lower.y + upper.y ) / 2;
    center.z = ( lower.z + upper.z ) / 2;

    ptf_onode_ref child;
    // WM call into ptf_octree_split always fails (invalid arguments!!) why is that??
    err = ptf_octree_split( octree, center, node, &child );
    if( err != peSuccess ) {
      return err;
    }
    return splitOctreeToFitCube( octree, child, cube, o_node );
  }
  else {
    *o_node = node;
    return peSuccess;
  }
}

/****************************************************************
 * End                                                          *
 ****************************************************************/

bool insert_spheroid( char const * fn, FILE * f, double center_x, double center_y, double center_z, 
  double majorRad, double squashFact, double textureU, double textureV )
{
  // open the ptf file
  ptf_file_handle ptfFile;
  ptf_char ptfFilename[ 1025 ];
  ptf_str_from_char( fn, ptfFilename, 1024 );
  ptf_err ptfErr = ptf_file_open( ptfFilename, ptf_open_update, &ptfFile );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to open ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }

  // acquire the octree root from the ptf file 
  ptf_file_header header;
  ptf_octree_handle octree;
  ptf_onode_ref octree_root;
  ptfErr = ptf_file_get_header( ptfFile, &header );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to get header from ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }
  ptfErr = ptf_octree_create( ptfFile, &header.range, &octree, 0, 0 );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to create octree from ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }
  ptfErr = ptf_octree_get_root( octree, &octree_root );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to get octree root from ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }

  // formula for calculating semi-minor radius = semi-major radius * ( 1 - 1/squash );
  if( squashFact <= 0 ) {
    fprintf(stderr, "Squash factor %f can not be smaller and equal to 0 \n", squashFact );
    return false;
  }
  double minorRad = majorRad * (1-(1/squashFact));
  if( minorRad <= 0 ) {
    fprintf(stderr, "Semi-minor radius %f can not be smaller and equal to 0\n", minorRad );
    return false;
  }

  // NOTE: we assume that the given center position is in a geo position 
  // calculate the min and max bounding box
  ptf_geopos gpMin, gpMax;
  gpMin.x = center_x - fabs( majorRad );
  gpMin.y = center_y - fabs( minorRad );
  gpMin.z = center_z - fabs( majorRad );
  gpMax.x = center_x + fabs( majorRad );
  gpMax.y = center_y + fabs( minorRad );
  gpMax.z = center_z + fabs( majorRad );

  // get the bounding box for the spheroid
  ptf_ocube spheroidCube;
  ptfErr = ptf_octree_coord_get_box_natural(gpMin, gpMax, &spheroidCube);
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to get the bounding box with err %d\n", ptfErr);
    return false;
  }

  // using the bounding box, split the octree down to the level and location 
  // where we want to place the spheroid
  ptf_onode_ref insertionNode;
  ptfErr = splitOctreeToFitCube( octree, octree_root, spheroidCube, &insertionNode );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to find the appropriate octree node to insert with err %d\n", ptfErr);
    return false;
  }

  // create and insert the spheroid to ptf file
  ptf_spheroid_handle spheroid = 0;
  ptfErr = ptf_spheroid_create( &spheroid, (ptf_char const *)"spheroid1" ); //WM what name should we used?
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to create spheroid object with err %d\n", ptfErr);
    return false;
  }
  ptf_spheroid_data data;
  ptfErr = ptf_spheroid_get_data(spheroid, &data);
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to get spheroid data with err %d\n", ptfErr);
    return false;
  }
  data.radius = majorRad;
  data.squash = squashFact;
  ptfErr = ptf_spheroid_set_data(spheroid, data);
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to set spheroid data with err %d\n", ptfErr);
    return false;
  }
  ptf_fixed_ref sphRef;
  ptfErr = ptf_spheroid_append_to_file(spheroid, ptfFile, &sphRef);
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to append spheroid to ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }

  // create content for spheroid  
  ptf_content_handle content;
  ptfErr = ptf_content_create( &content );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to create content for spheroid with err %d\n", ptfErr);
    return false;
  }
  // check whether the content exists in ptf file or not
  ptf_fixed_ref contentRef;
  bool updatingContents = false;
  
  if( insertionNode.content_index != 0 ) {
    contentRef.index = insertionNode.content_index;
    ptfErr = ptf_content_read_from_file( content, ptfFile, contentRef );
    if( ptfErr != peSuccess ) {
      fprintf(stderr, "Fail to read content from ptf file %s with err %d\n", fn, ptfErr);
      return false;
    }
    updatingContents = true;
  }

  // add a new spheroid reference inside the content
  ptf_record_ref record;
  ptf_32 contentIndex;
  memset(&record, 0, sizeof(record));
  record.type = ptSpheroidData;
  record.index = sphRef.index;
  ptfErr = ptf_content_add_reference( content, record, &contentIndex );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to add new spheroid reference with err %d\n", ptfErr);
    return false;
  }

  // add a new spheroid instance inside the content
  ptf_32 instanceIndex;
  ptf_geopos pos;
  pos.x = center_x;
  pos.y = center_y;
  pos.z = center_z;
  ptf_ori ori; // no rotation
  ori.x = 0;  
  ori.y = 0;
  ori.z = 0;
  ori.w = 1;
  ptfErr = ptf_content_add_reference_instance( content, contentIndex, pos, ori, &instanceIndex );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to add new spheroid instance with err %d\n", ptfErr);
    return false;
  }

  if( updatingContents ) {
    // update the content in ptf file
    ptfErr = ptf_content_update_to_file( content, ptfFile, contentRef );
    if( ptfErr != peSuccess ) {
      fprintf(stderr, "Fail to update the content in ptf file %s with err %d\n", fn, ptfErr);
      return false;
    }
  }
  else {
    // write the new content to ptf file 
    ptfErr = ptf_content_append_to_file( content, ptfFile, &contentRef );
    if( ptfErr != peSuccess ) {
      fprintf(stderr, "Fail to append a new content to ptf file %s with err %d\n", fn, ptfErr);
      return false;
    }

    // update the content index
    insertionNode.content_index = contentRef.index;
    ptfErr = ptf_octree_update_content_index( octree, insertionNode ); 
    if( ptfErr != peSuccess ) {
      fprintf(stderr, "Fail to update spheroid content index in ptf file %s with err %d\n", fn, ptfErr);
      return false;
    }
  }

  // flush the ptf octree
  ptfErr = ptf_octree_close( octree );
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to flush octree to ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }

  // close the ptf file
  ptfErr = ptf_file_close(ptfFile);
  if( ptfErr != peSuccess ) {
    fprintf(stderr, "Fail to close ptf file %s with err %d\n", fn, ptfErr);
    return false;
  }
  return true;
}

int main(int argc, char * argv[]) {
  int cmd = 0;
  if (argc > 2) {
    if (!strcmp(argv[1], "dump")) {
      cmd = 0;
    }
    else if (!strcmp(argv[1], "dumptree")) {
      cmd = 1;
    }
    else if (!strcmp(argv[1], "add")) {
      cmd = 2;
    }
    else if (!strcmp(argv[1], "remove")) {
      cmd = 3;
    }
    else if (!strcmp(argv[1], "replace")) {
      cmd = 4;
    }
    else if (!strcmp(argv[1], "extract")) {
      cmd = 5;
    }
    else if (!strcmp(argv[1], "dumpcontent")) {
      cmd = 6;
    }
    else if (!strcmp(argv[1], "dumpmesh")) {
      cmd = 7;
    }
    else if( !strcmp(argv[1], "insertspheroid")) {
      cmd = 8;
    }
    //  should have a "compact" command, too
    else {
      goto usage;
    }
    ++argv;
    --argc;
  }
  else {
usage:
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "ptf dump filename.ptf\n");
    fprintf(stderr, "ptf dumptree filename.ptf\n");
    fprintf(stderr, "ptf add filename.ptf type datafile\n");
    fprintf(stderr, "ptf remove filename.ptf type index\n");
    fprintf(stderr, "ptf replace filename.ptf type index datafile\n");
    fprintf(stderr, "ptf extract filename.ptf type index datafile\n");
    fprintf(stderr, "ptf dumpcontent filename.ptf [index]\n");
    fprintf(stderr, "ptf dumpmesh filename.ptf [index]\n");
    fprintf(stderr, "ptf insertspheroid filename.ptf centerX centerY centerZ majorRadius squashFactor textureU textureV\n");
    return 1;
  }
  ++argv;
  --argc;
  FILE * f = fopen(argv[0], "rb+");
  if (!f) {
    char dir[1024];
    getcwd(dir, 1024);
    dir[1023] = 0;
    fprintf(stderr, "%s: can't open\n", argv[0]);
    fprintf(stderr, "current dir: %s\n", dir);
    return 2;
  }
  switch (cmd) {
    case 0:
      if (argc != 1) {
        goto usage;
      }
      dump_header(argv[0], f);
      dump_meta(argv[0], hdr, f);
      break;
    case 1:
      dump_header(argv[0], f);
      dump_tree(argv[0], hdr, f);
      break;
    case 2:
      if (argc != 3) {
        goto usage;
      }
      add_block(argv[0], f, argv[1], argv[2]);
      break;
    case 3:
      if (argc != 3) {
        goto usage;
      }
      remove_block(argv[0], f, argv[1], argv[2]);
      break;
    case 4:
      if (argc != 4) {
        goto usage;
      }
//      replace_block(argv[0], f, argv[1], argv[2], argv[3]);
fprintf(stderr, "not implemented\n");
      break;
    case 5:
      if (argc != 4) {
        goto usage;
      }
//      extract_block(argv[0], f, argv[1], argv[2], argv[3]);
fprintf(stderr, "not implemented\n");
      break;
    case 6:
      if (argc > 2) {
        goto usage;
      }
      if (argv[1]) {
        dump_content(argv[0], f, cvt_type(argv[1]));
      }
      else {
        dump_all_content(argv[0], f);
      }
      break;
    case 7:
      if (argc > 2) {
        goto usage;
      }
      if (argv[1]) {
        dump_mesh(argv[0], f, cvt_type(argv[1]));
      }
      else {
        dump_all_mesh(argv[0], f);
      }
      break;
    case 8:
      if(argc != 8) {
        goto usage;
      }
      if( !insert_spheroid(/*ptf_filename*/argv[0], f,
                           /*centerX*/     strtod( argv[1], NULL ), 
                           /*centerY*/     strtod( argv[2], NULL ),
                           /*centerZ*/     strtod( argv[3], NULL ),
                           /*majorRadius*/ strtod( argv[4], NULL ), 
                           /*squashFactor*/strtod( argv[5], NULL ),
                           /*textureU*/    strtod( argv[6], NULL ),
                           /*textureV*/    strtod( argv[7], NULL )) )
      {
        fprintf(stderr, "insert spheroid command fails" );
        exit(1);
      }
      break;
    default:
      fprintf(stderr, "unknown command: %s (%d)\n", argv[-1], cmd);
      exit(1);
  }
  return 0;
}
