
#if !defined(ptf_material_h)
//! \internal header guard
#define ptf_material_h

//! \file ptf_material.h
//!
//! Materials are complex entities, because they must represent a wide 
//! variety of possible ways of describing how a surface looks. In some 
//! applications, you may even have different descriptions for different 
//! sensors, such as radar, IR, UV or visible spectrum.
//!
//! Materials are stored in a material archive. References to materials 
//! are stored as record references for the archive (see ptf_fixed_ref) 
//! plus an index into the archive. An index within the archive is semi 
//! permanent, in that removing a material will not re-number the other 
//! materials.
//!
//! Material parameters are of various type (float, texture map, etc) which 
//! may in turn need description. Those additional pieces of data are 
//! stored as child parameters of the material parameter. Such recursion 
//! can go as deep as necessary. However, the defined material parameter 
//! types are currently:
//!
//! <table>
//! <tr><th>Parent</th><th>Parameter</th><th>Type</th><th>Notes</th></tr>
//! <tr><td>pmtMaterialArchive</td><td>&lt;material name&gt;</td><td>pmtMaterial</td>
//! <td>An archive can contain multiple child materials, referenced by index.</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>&lt;any&gt;</td>
//! <td>Well-defined parameter names include 'diffuse' for the diffuse material color.</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtFloatTuple</td><td>A constant value</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtUInt32Tuple</td><td>A constant value</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtString</td><td>Some annotation</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtTextureMapExplicit</td><td>A texture map using explicit UV coordinates from the geometry</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtTextureMapImplicit</td><td>A texture map using implicit UV coordinates generated from a plane.</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtTextureMapGridded</td><td>A set of texture maps that together map an area</td></tr>
//! <tr><td>pmtMaterial</td><td>&lt;parameter name&gt;</td><td>pmtMaterialMap</td><td>A texture map which selects between different "kind" textures</td></tr>
//! <tr><td>pmtTextureMapExplicit</td><td>texture_data</td><td>pmtRecordRef</td><td>Reference to a texture map, either a TextureDataRef, or an inline TextureData</tr>
//! <tr><td>pmtTextureMapExplicit</td><td>texture_map_name</td><td>pmtString</td><td>The name of this texture map (advisory only)</td></tr>
//! <tr><td>pmtTextureMapExplicit</td><td>physical_scale_u</td><td>pmtFloatTuple</td><td>A single float, measuring the physical extent (in meters) of this texture map, in the U direction</td></tr>
//! <tr><td>pmtTextureMapExplicit</td><td>physical_scale_v</td><td>pmtFloatTuple</td><td>A single float, measuring the physical extent (in meters) of this texture map, in the V direction</td></tr>
//! <tr><td>pmtTextureMapImplicit</td><td>texture_map</td><td>pmtTextureMapExplicit</td><td>The underlying texture data</td></tr>
//! <tr><td>pmtTextureMapImplicit</td><td>reference_pos</td><td>pmtDoubleTuple</td><td>The geopos at which U,V are 0,0</td></tr>
//! <tr><td>pmtTextureMapImplicit</td><td>u_vector</td><td>pmtFloatTuple</td><td>This vector, dot the geopos minus the reference pos, equals U texture coordinate</td></tr>
//! <tr><td>pmtTextureMapImplicit</td><td>v_vector</td><td>pmtFloatTuple</td><td>This vector, dot the geopos minus the reference pos, equals V texture coordiante</td></tr>
//! <tr><td>pmtTextureMapImplicit</td><td>clamp_behavior</td><td>pmtString</td><td>Single character: w for wrap, e for edge clamp, b for border clamp</td></tr>
//! <tr><td>pmtTextureMapGridded</td><td>texture_map</td><td>pmtTextureMapImplicit</td><td>Multiple children of this name go together to make up the full texture. Each child should be clamping U/V.</td></tr>
//! <tr><td>pmtMaterialMap</td><td>material_map</td><td>pmtTextureMapImplicit</td><td>The material mapping texture for a multi-map. Each pixel index maps to one of the texture maps.</td></tr>
//! <tr><td>pmtMaterialMap</td><td>texture_map</td><td>pmtTextureMapExplicit</td><td>Each instance defines one texture map for a given material kind</td></tr>
//! </table>

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library
//! @{
//! \copydoc ptf_material.h


//! Material descriptions are identified by material handles. 
//! These handles exist separately from the PTF files, but can
//! only be created in a material archive, which can read and 
//! write materials from/to PTF files.
typedef struct ptf_material_t *ptf_material_handle;

//! A material archive is a collection of material handles.
//! The archive can be serialized to and from PTF files.
typedef struct ptf_material_archive_t *ptf_material_archive;

//! A material parameter is any variable that can be stored with 
//! a material, to describe its rendering or behavior.
typedef struct ptf_material_parameter_t *ptf_material_parameter;

enum {
  PTF_CURRENT_MATERIAL_VERSION = 1,   //!<  The version of material actually read by this library
  PTF_MIN_MATERIAL_VERSION = 1,       //!<  The version of material needed to read what's written by this library
};

//! Create a material archive.
//! \param o_arc The archive handle is returned here.
//! \return ptf_err status
PTF_API ptf_err ptf_material_archive_create(ptf_material_archive *o_arc);

//! Destroy a material archive, de-allocating all data within.
//! \param arc The material archive to deallocate.
//! \return ptf_err status
PTF_API ptf_err ptf_material_archive_destroy(ptf_material_archive arc);

//! Write a material archive to a new record in a file.
//! \param arc The archive.
//! \param file The file to write to.
//! \param ref The record reference of the written archive.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_append_to_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref *o_ref);

//! Update a material archive to an existing record in a file.
//! \param arc The archive.
//! \param file The file to write to.
//! \param ref The (existing) record reference to update.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_update_to_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Read the contents of an existing material archive from disk.
//! Whatever data was previously in the in-memory archive will be replaced 
//! by the data from disk.
//! \param arc The material archive to set to the contents of the file record.
//! \param file The open PTF file to read from.
//! \param ref The record to read from the file.
//! \param o_version The actual version of the archive is returned here. If 
//! this is larger than PTF_CURRENT_MATERIAL_VERSION, then saving this archive 
//! back out to disk may lead to data loss for clients using later versions 
//! of the library.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_read_from_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref const &ref, ptf_16 * o_version);

//! Create a material in an archive. Return the handle to the material.
//! \param arc The archive to receive the material.
//! \param name The name to give the material. This name must be unique 
//! within the archive.
//! \param o_mat The created material handle is returned here.
//! \return ptf_err status.
//! \note Well-known properties of materials include: <ul>
//! <li>"diffuse" -- type pmtTextureMapExplicit, diffuse color value map</li>
//! <li>"shadername" -- type pmtString, name of shader to render material</li>
//! <li>"material_map" -- type pmtMaterialMap, a coverage map where each pixel maps to a separate material</li>
//! </ul>
//! The expected hierarchy of properties is documented on the \ref terrainmaterial page.
PTF_API ptf_err ptf_material_archive_create_material(ptf_material_archive arc, 
    ptf_char const *name, ptf_material_handle *o_mat);

//! Remove a material from an archive. This invalidates the material 
//! handle.
//! \param mtl The material to remove from the archive, and destroy.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_remove(ptf_material_handle mtl);

//! Find a material handle given the material name.
//! \param arc The archive to look in.
//! \param name The name to look for.
//! \param o_mat The found material handle is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_find_material(ptf_material_archive arc,
    ptf_char const *name, ptf_material_handle *o_mat);

//! Get a material handle given an index into the archive. The indices are
//! semi-stable; the archive does not compact on removal, but empty indicies 
//! may be re-used for new materials. A de-allocated material will return 
//! a NULL handle, but no error.
//! \param arc The archive.
//! \param index The index of the material to return. Must be between 0 (inclusive) 
//! and ptf_material_archive_count_materials() (exclusive).
//! \param o_mat The found material handle is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_get_material(ptf_material_archive arc,
    ptf_32 index, ptf_material_handle *o_mat);

//! Count the number of materials in the file. Note that some of the materials 
//! may be de-allocated, and return NULL.
//! \param arc The archive.
//! \param o_count The number of materials is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_archive_count_materials(ptf_material_archive arc,
    ptf_32 * o_count);

//! Retrieve name of the given material.
//! \param mat The material.
//! \param m_name A pointer to the name of the parameter will be returned here.
//! This pointer is valid until the material or archive is next mutated.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_get_name(ptf_material_handle mat,
    ptf_char const **m_name);

//! Count the number of parameters for a material.
//! \param mat The material.
//! \param o_count The number of parameters.
PTF_API ptf_err ptf_material_count_parameters(ptf_material_handle mat,
    ptf_32 * o_count);

//! Get a handle for a specific parameter (found by index) on a given material.
//! \param mat The material to look for parameters in.
//! \param index The index (nth) parameter to get.
//! \param o_param The parameter will be returned here.
//! \return ptf_err status.
//! \note Parameters at the top level of the material are specific material 
//! quantities like "diffuse" for diffuse color, or "power" for specular power.
//! The value of those parameters is expressed through each parameter type 
//! and could be constant, a texture map, or some more complex function.
PTF_API ptf_err ptf_material_get_parameter(ptf_material_handle mat,
    ptf_32 index, ptf_material_parameter *o_param);

//! Find the first parameter with a name that matches the provided string. 
//! Note that, if multiple parameters with the same name are stored, only 
//! one of them will be returned. use ptf_material_get_parameter() to get 
//! all parameters.
//! \param mat The material to search for parameters in.
//! \param name The name of the parameter to find.
//! \param o_param The parameter will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_material_find_parameter(ptf_material_handle mat,
    ptf_char const *name, ptf_material_parameter *o_param);

//! Retrieve type and name information about the given parameter.
//! \param param The parameter to inspect.
//! \param o_name A pointer to the name of the parameter will be returned here.
//! This pointer is valid until the next mutation operation on the material 
//! archive.
//! \param o_type The type of the parameter is returned here.
//! \param o_size The size of the parameter data (if any) is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_get_info(ptf_material_parameter param,
    ptf_char const **o_name, ptf_material_type *o_type, size_t *o_size);

//! Add a parameter to a given material description.
//! \param mat The material to add the parameter to.
//! \param name The name of the material to add (must be non-empty).
//! \param type The type of the material parameter.
//! \param o_param The new parameter is returned here.
//! \note The material archive owns the parameter, so you do not need to dispose it.
PTF_API ptf_err ptf_material_add_parameter(ptf_material_handle mat,
    ptf_char const *name, ptf_material_type type, ptf_material_parameter *o_param);

//! Most parameters have sub-parameters to specify various information, 
//! such as UV tiling mode, or texture coordinate generation planes. 
//! You can add such a parameter to an existin parameter using this function.
//! The set of legal sub-parameters varies by head parameter type.
//! \param param The parameter to add a child parameter to.
//! \param name The name of the new parameter to add.
//! \param type The type of the new parameter.
//! \param o_param The new parameter is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_add_parameter(ptf_material_parameter param,
    ptf_char const *name, ptf_material_type type, ptf_material_parameter *o_param);

//! Remove an existing parameter from its parent (parameter or material).
//! The parameter handle becomes invalid after this function.
//! \param param The parameter to remove.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_remove(ptf_material_parameter param);

//! Get the value of a parameter. You provide the size of the buffer. 
//! If the size is not sufficient, the function will store the needed size 
//! in the size argument, but not overwrite the end of your buffer.
//! \param param The parameter to get parameter data from.
//! \param data A pointer to the storage for the parameter data to copy out.
//! \param ioSize in: the size of the buffer pointed at by 'data'.
//! out: the actual size of the material data.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_get_value(ptf_material_parameter param,
    void *data, size_t *ioSize);

//! Set the value of a parameter. The value should be of the correct type 
//! for the parameter, but is not checked.
//! Parameters with child parameters cannot also have values.
//! \param param The parameter to set the value of.
//! \param data A pointer to the value of the parameter to copy in.
//! \param size The number of bytes to copy into the parameter.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_set_value(ptf_material_parameter param,
    void const *data, size_t size);

//! Figure out how many parameters are children of this parameter.
//! \param param The parent parameter to enumerate.
//! \param o_count The number of child parameters will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_count_parameters(ptf_material_parameter param,
    ptf_32 *o_count);

//! Find a child parameter by name. If there are multiple child parameters 
//! with the same name, only one will be found.
//! \param param The parameter to look for a child in.
//! \param name The name of the parameter to find.
//! \param o_param The found child parameter will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_find_parameter(ptf_material_parameter param,
    ptf_char const *name, ptf_material_parameter *o_param);

//! Get a child parameter of a parameter by index.
//! \param param The parent parameter.
//! \param index The index of the child parameter to get.
//! \param o_h The parameter child will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_parameter_get_parameter(ptf_material_parameter param,
    ptf_32 index, ptf_material_parameter *o_h);

//! @}
#endif  //  ptf_material_h
