#if !defined(ptf_octree_h)
//! \internal header guard
#define ptf_octree_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif

//! \addtogroup PtfGeometry The PTF Geometry Library


//! The type used to identify an octree instance for a file. There will 
//! be a 1:1 relation between the octree instance and the file, although 
//! the instance may create multiple octree chunks in the file.
//! \note Carefully constructed ill formed files may cause the octree 
//! code to fail validation, and either enter an infinite loop, or crash 
//! the program.
typedef struct ptf_octree_t * ptf_octree_handle;

//! References to a specific octree node are passed as ptf_onode_ref.
//! \note This structure is only used in the interface to the octree 
//! code; it is not used as-is on disk and is not a normative data 
//! structure.
struct ptf_onode_ref {
  ptf_32      file_index;     //!<  When this is non-0, you have to go open another file!
  ptf_32      octree_index;   //!<  Index to record of type ptOctreeNode.
  ptf_16      node_index;     //!<  Index within the given record.
  ptf_16      node_level;     //!<  Level within the per-block octree (how far down it's split).
  ptf_32      content_index;  //!<  Index of content record for this node.
  ptf_ocube   extent;         //!<  What extent of space is covered by this particular node.
};

//! Create the root octree of the file, and enable it for 
//! traversal or update. This does an implicit read (and modifications 
//! write) on the file, so no more than one octree handle should exist 
//! for a given file at any one time. If you create more than one, the 
//! same handle will be returned, and reference counted.
//! \param file The open file to load an octree from.
//! \param io_range The extent that will be created for the octree.
//! The range must describe a single cube (at some level) of the 
//! global octree. If the octree already exists in the file, the 
//! actual range will be returned here. The cube must be aligned to 
//! a natural boundary of its size, or half of that boundary (for 
//! offset octal roots).
//! \param o_handle The open octree will be returned here.
//! \param max_depth Set the maximum depth allowed per node record in 
//! the ptf_file. This trades depth vs width versus number of records 
//! used to represent a given octree. Set to 0 to get a sensible default. 
//! This value is per invocation, and not stored within the file. The 
//! value will be clamped to within legal range.
//! \param max_nodes Set the maximum number of nodes allowed per record.
//! Set to 0 to get a sensible default. This value is per invocation, 
//! and not stored within the file. The value will be clamped to 
//! within legal range. Its use is to limit the maximum size of each 
//! individual octree node record for very small host machines.
//! \return ptf_err status.
//! \note If the octree is split beyond the max_depth and max_nodes 
//! parameters, new octree record_refs are automatically generated in 
//! the ptf_file, i e, the logical octree within a file can span more 
//! than one physical octree type record, each of which contains some 
//! number of octree nodes.
PTF_API ptf_err ptf_octree_create(ptf_file_handle file, ptf_ocube * io_range, 
    ptf_octree_handle * o_handle, ptf_16 max_depth, ptf_16 max_nodes);

//! Close an open octree. Warning! This is not done automatically when 
//! closing a ptf_file_handle; you have to do it manually (as the 
//! octree API sits on top of the ptf_file API).
//! \param oh The octree handle to close.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_close(ptf_octree_handle oh);

//! Resolve a point to an octree node, as far as the tree goes. 
//! Does not add new nodes.
//! \param oh The open octree handle.
//! \param pos The position to classify (using global octree extents).
//! \param o_ref The classified node will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_resolve(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_onode_ref * o_ref);

//! Resolve a point to an octree node to the specified level. 
//! Does not add new nodes.  If the specified level does not exist,
//! then return the deepest level node that does exist.
//! \param oh The open octree handle.
//! \param level The octree level to stop resolving at. This is measured 
//! in the global level space (from the outmost cube of the logical top 
//! octree), not in levels within this particular file.
//! \param pos The position to classify (using global octree extents).
//! \param o_ref The classified node will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_resolve_to_level(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_16 level, ptf_onode_ref * o_ref);

//! Resolve a point to an octree node child, down a single level.
//! Can add new nodes to the file (which may, worst case, add a whole 
//! new node chunk).
//! \param oh The open octree handle.
//! \param pos The position to resolve (using global octree extents).
//! \param parent The parent to be split. Start out with ptf_octree_get_root().
//! \param o_ref The classified node will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_split(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_onode_ref const & parent, ptf_onode_ref * o_ref);

//! Fill in an array of children of a given node.
//! \param oh The open octree handle.
//! \param parent The parent node.
//! \param children The array of children will be returned here.
//! If a child is not populated (i e, the parent is a terminal node), the 
//! node_level value of that child will be set to 0. For all valid children, 
//! node_level is greater than 0 (as 0 means the root, which is not a child).
//! \return ptf_err status.
//! \note The children array must have room for 8 entries 
PTF_API ptf_err ptf_onode_get_children(ptf_octree_handle oh, ptf_onode_ref const & parent,
    ptf_onode_ref * children);

//! Set the "content" index for a given node.
//! This only updates the tree data, it does not do anything to the 
//! records of type ptOctreeContent (which contain references to meshes etc.)
//! \param oh The open octree handle.
//! \param ref The node ref with the updated "content_index" value. All other 
//! values should stay the same.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_update_content_index(ptf_octree_handle oh,
    ptf_onode_ref const & ref);

//! Get the reference for the root node of the octree.
//! \param oh The open octree handle.
//! \param o_root The root reference is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_get_root(ptf_octree_handle oh, ptf_onode_ref * o_root);

//! Add a reference to an external file into the current octree. It is an error 
//! for content to be stored below this reference -- you must delete such content 
//! first.
//! \param oh The octree to add the external file reference to.
//! \param ref_file The file to add a reference to. The area covered by the file's 
//! header extent will be replaced. It is an error for the original octree to 
//! contain contents that overlaps with the referenced file (at a lower octree 
//! level -- shared content added at a higher level is still allowed).
//! \return ptf_err status.
PTF_API ptf_err ptf_octree_set_reference(ptf_octree_handle oh, ptf_file_handle ref_file);

//! Convert a given point to a set of bounding box coordinates. The box will 
//! be sized based on the 'level' argument (which should be between 0 and 30).
//! Level 0 is the top level of the octree (range [0x00000000,0x80000000)).
//! \param pos The position to convert.
//! \param level The size of the bounding cube to return (range 0 through 30).
//! \param o_box The bounding box.
PTF_API ptf_err ptf_octree_coord_get_box(ptf_geopos const & pos, ptf_16 level, 
    ptf_ocube * o_box);

//! Convert a given bounding box to a set of bounding box coordinates. The box 
//! will be sized based on the 'natural' level for the arguments (which should 
//! be between 0 and 30). This box may be off-by-half compared to the boxes 
//! stored in the actual octree, and thus represent 8 different sub-cubes.
//! \param pos The position to convert.
//! \param o_box The bounding box.
PTF_API ptf_err ptf_octree_coord_get_box_natural(ptf_geopos const &min, 
    ptf_geopos const &max, ptf_ocube * o_box);

//! Convert a given bounding box to the coordinates for the lower-left-yon and 
//! upper-right-hither corners (minimum and maximum values).
//! \param box The box to convert.
//! \param o_lower The coordinates for the lower left yon point of the box are returned here.
//! \param o_upper The coordinates for the upper right hither point of the box are returned here.
//! \note X right, Y up, Z hither.
PTF_API ptf_err ptf_octree_box_get_coords(ptf_ocube const & box, ptf_geopos * o_lower, 
    ptf_geopos * o_upper);

//! Set the coordinates for the given box.
//! \param o_lower The new coordinates for the lower-left-yon point of the box.
//! \param o_upper The new coordinates for the upper-right-hither point of the box.
//! \param box The box to modify.
//! \note X right, Y up, Z hither.
PTF_API ptf_err ptf_octree_box_set_coords(ptf_geopos const & o_lower, ptf_geopos const & o_upper,
    ptf_ocube * box);

//! Convert from meters into coordinate units used by the octree.
//! The result will be truncated.
//! \param meters The meter value to convert. Must not be negative, or larger than an octree is 
//! capable of handling (i e, slightly larger than Earth with atmosphere).
//! \param geo_units The converted octreecoord value.
//! \note ptf_geopos is in meters. The octreecoord starts with 0 at the lower boundary of the 
//! octree (PTF_MIN_GEOPOS), so 0 for "meters" means 0x40000000 for geo_units.
PTF_API ptf_err ptf_meters_to_octreecoord( double meters, ptf_32 * geo_units );

//! Convert from coordinate units used by the octree to meters
//! \param geo_units The converted meter value (in octree units).
//! \param meters The octree-unit value converted to meters is returned here.
PTF_API ptf_err ptf_octreecoord_to_meters( ptf_32 geo_units, double * meters );

//! Get the length of an ocube side at the given level.
//! \param level The octree cube level.
//! \param side The length of one side of a cube at given level.
//! \note The level 0 box side is 0x80000000
PTF_API ptf_err ptf_level_to_box_side( ptf_32 level, ptf_32 * side );

//! Get the octree level given the length of the cube edge
//! \param side The length of one side of a cube.
//! \param level The octree cube level.
//! \note The level 0 box side is 0x80000000
PTF_API ptf_err ptf_box_side_to_level( ptf_32 side , ptf_32 * level );

//! Get information about a given octree cube.
//! \param cube The cube to get information about.
//! \param o_level The octree level this cube represents
//! \param o_side The side length of the cube
//! \param o_offset Whether the cube is offset-by-half or not along 
//! each axis: 1 == x, 2 == y, 4 == z, returned as a bit-mask.
//! \return ptf_err status.
//! \note This function will return an error if the cube is not 
//! suitable for octree use.
PTF_API ptf_err ptf_octree_get_box_info(ptf_ocube const &box, ptf_32 *o_level, ptf_32 *o_side, ptf_32 *o_offset);

//! Test for equality between ocubes
//! \param a Left side of the operator.
//! \param b Right side of the operator.
//! \return true if they are equal in size and location.
PTF_API bool ptf_ocube_equals_ocube(ptf_ocube const &a, ptf_ocube const &b);

//! Test for containment between ocubes
//! \param abig The ocube that should contain the other.
//! \param bsmall The ocube that should be contained in the first.
//! \return true if the small ocube does not extend past the bounds of the big.
PTF_API bool ptf_ocube_contains_ocube(ptf_ocube const &abig, ptf_ocube const &bsmall);

#endif  //  ptf_octree_h
