#if !defined(ptf_texture_h)
//! \internal header guard
#define ptf_texture_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library

//! Texture data is identified by a ptf_texture_handle.
//! These handles just represent the data, and can be 
//! loaded from and written to PTF files, but do not 
//! remember what file they came from.
typedef struct ptf_texture_t *ptf_texture_handle;

//! Allocate a texture object. This object exists in memory until you destroy 
//! it. There are functions to read data into the object from a PTF file, and 
//! to write the data in memory to a PTF file, but the object does not retain 
//! any knowledge about the file.
//! A texture object contains some number of surfaces (1 for a typical texture 
//! image, 6 for a cube map). Each surface, in turn, contain some number of 
//! MIP map level images, where each level is a decimated version of the previous 
//! level. Additionally, each image may contain data for a border.
//! \see ptf_texture_data
//! \param o_tex The allocated texture object will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_create(ptf_texture_handle *o_tex);

//! De-allocate a texture object from memory.
//! After this call, all references to tex will be invalid.
//! \param tex The texture object to de-allocate.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_destroy(ptf_texture_handle tex);

//! Make room for a given number of surfaces with a given number of MIP maps, 
//! using given dimensions. Actual memory for the data won't be allocated until 
//! ptf_texture_set_data is called. However, any previously allocated data will 
//! be cleared, and the texture object returned to an empty state.
//! This function must be called after ptf_texture_create if you intend to add 
//! data to the texture (rather than read data from a file).
//! \note You must specify at least one dimension, surface and mip map level.
//! \param tex The texture to specify the structure of.
//! \param header The header containing the specification. Most fields matter, 
//! except for the data_offset and data_size fields.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_resize(ptf_texture_handle tex, ptf_texture_data const * header);

//! Put actual data for a single MIP map level image for a single surface into the 
//! texture, replacing any previous data that might have existed in that level for 
//! that surface. The library does not verify that the data is in a correct format 
//! for the header; it just stores the data. Note that the alignment parameter of 
//! the header applies to each scanline for raw pixel formats, but to each MIP map 
//! level image for compressed formats. Also, the data set here will be padded up 
//! to that alignment automatically when writing out to file.
//! \param tex The texture object to copy data into.
//! \param surface The destination surface. Typically 0, unless you're storing data 
//! for a cube map. Must be less than the number of surfaces specicied in the header.
//! \param mipmap The MIP map level destination image. Image 0 is the largest image.
//! Must be less than the number of mip levels specified in the header.
//! \param data A pointer to the data to copy into the texture object for the specified 
//! surface and mip map level.
//! \param size The size of the data to copy into the texture object.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_set_data(ptf_texture_handle tex, ptf_8 surface, ptf_8 mipmap,
    void const * data, size_t size);

//! Get information about the data stored in the texture object.
//! \param tex The texture object to return information about.
//! \param o_header The texture information.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_get_info(ptf_texture_handle tex, ptf_texture_data *o_header);

//! Retrieve a pointer to the data stored for a given texture surface and MIP level. 
//! \param tex The texture object to get data from.
//! \param surface The surface index to get data from (typically 0).
//! \param mipmap The mip map level to get data from, where 0 is the largest image.
//! \param o_data A pointer to the stored data will be returned here.
//! \param o_size The number of bytes of stored data will be returned here.
//! Pay attention to the alignment parameter of the header, which tells the minimum 
//! alignment of the start of each scanline, for raw pixel formats.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_get_data(ptf_texture_handle tex, ptf_8 surface, ptf_8 mipmap,
    void const **o_data, size_t *o_size);

//! Given an existing texture object, remove any currently stored data, and read data 
//! from the given PTF file record into the texture object.
//! \param tex The texture object to read into from file.
//! \param file The open PTF file to read from.
//! \param ref The record (of type ptTextureData) to read from.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_read_from_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given an existing texture object that has data in it, write that texture object 
//! to a new record in an open PTF file, and return a reference to that record. This 
//! does not modify the texture object.
//! \param tex The texture object to write to the file.
//! \param file The open PTF file to append to.
//! \param o_ref The record reference will be returned here. The record will be of 
//! type ptTextureData.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_append_to_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref *o_fr);

//! Given an existing texture object that has data in it, write that texture object 
//! to an existing record in an open PTF file, and replace the previous contents of 
//! that record. This does not modify the texture object.
//! \param tex The texture object to write to the file.
//! \param file The open PTF file to append to.
//! \param ref The record reference of the ptTextureData record to replace.
//! \return ptf_err status.
PTF_API ptf_err ptf_texture_update_to_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref const &ref);

#endif  //  ptf_texture_h
