#include "ptf.h"
#include "ptf_collection.h"
#include "ptf_str.h"
#include "ptf_octree.h"

#include "mem_util.h"

#include <assert.h>
#include <string.h>

#include <vector>
#include <queue>




/* Reminder:
 *
 * The actual data type referred to by this collection is specified in the 
 * header of the collection. It could be mesh data (ptMeshData) or something 
 * totally different.
 */

struct ptf_collection_on_disk_node {
  ptf_32            cld_and_parent;   //!< This node is child X of parent Y, where X is first 3 bits, Y is last 29
  ptf_remote_ref    mesh;             //!< The mesh this node refers to
  float             feature_size;     //!< The feature size of the mesh being referenced
};

struct ptf_internal_crec_t {
  ptf_crec_t crec;
  ptf_32 id;
};

struct ptf_internal_disk_node {
  ptf_collection_on_disk_node disk_node;
  ptf_32                      left, bottom, yon;
};

class ptf_collection_t {
  public:
    ptf_collection_t(ptf_type type);
    ~ptf_collection_t();

    ptf_err marshal_data(void **o_data, size_t *o_size);

    ptf_collection_header           header;
    ptf_internal_disk_node          root_node;
    ptf_ocube                       overall_bounds;
    bool                            root_set;
    std::vector<ptf_crec_t>         records;
    std::vector<ptf_collection_on_disk_node> child_nodes;
    std::vector<char>               outputblock;
    std::vector<ptf_record_ref>     related_;
};

ptf_collection_t::ptf_collection_t(ptf_type type)
{
  root_node.bottom = root_node.left = root_node.yon = 0;
  root_node.disk_node.cld_and_parent = 0;
  root_node.disk_node.feature_size = 0.0;
  root_node.disk_node.mesh.file = 0;
  root_node.disk_node.mesh.index = 0;
  root_set = false;
  child_nodes.clear();
  records.clear();
  header.type = type;
  header.related = 0;
  header.related_offset = sizeof(header);
  header.bottom = 0;
  header.data_offset = 0; //  depends on number of related
  header.left = 0;
  header.num_nodes = 0;
  header.octree_level = 0;
  header.yon = 0;
  memset(header.name, 0, sizeof(header.name));
}

ptf_collection_t::~ptf_collection_t()
{
}

ptf_err ptf_collection_t::marshal_data(void **o_data, size_t *o_size)
{
  assert( root_set );
  assert((1 + child_nodes.size() ) == header.num_nodes);
  try {
    if(header.num_nodes < 1) {
      return peInvalidArguments;
    }
    std::vector<char> & output = outputblock;
    output.resize(0);
    size_t siz = sizeof(ptf_collection_header);
    header.related = (ptf_32)related_.size();
    header.related_offset = (ptf_32)siz;
    siz += sizeof(ptf_record_ref) * header.related;
    header.data_offset = (ptf_32)siz;
    siz += sizeof(ptf_collection_on_disk_node) * header.num_nodes;
    output.resize(siz);

    size_t pos = 0;

    //  create header
    memcpy(&output[pos], &header, sizeof(ptf_collection_header));
    pos += sizeof(ptf_collection_header);

    //  store related records
    if (output.size() > 0 && related_.size() > 0) {
      memcpy(&output[pos], &related_[0], sizeof(ptf_record_ref) * header.related);
      pos += sizeof(ptf_record_ref) * header.related;
    }
    // Keep track of where each parent ends up in the records list
    std::vector<ptf_32> index_locations;
    index_locations.resize(header.num_nodes+1);
    index_locations[0] = 0;

    std::queue<ptf_32> write_queue;
    memcpy(&output[pos], &(root_node.disk_node), sizeof(ptf_collection_on_disk_node));
    pos += sizeof(ptf_collection_on_disk_node);
    write_queue.push(1);
    index_locations[1] = 1;

    // Write out data in breadth-first order
    ptf_32 size = (ptf_32)child_nodes.size();
    ptf_32 queue_location = 1;
    while(!write_queue.empty()) {
      ptf_32 parent = write_queue.front();
      write_queue.pop();
      for(ptf_32 i = 0; i < size; ++i) {
        if((child_nodes[i].cld_and_parent & ((1UL<<29)-1)) == parent) {
          ptf_collection_on_disk_node temp = child_nodes[i];

          // Rebuild the parent index
          ptf_32 child_idx = temp.cld_and_parent >> 29;
          ptf_32 parent_idx = index_locations[parent];
          temp.cld_and_parent = ((child_idx & 7) << 29) | (parent_idx & ((1UL<<29)-1));

          memcpy(&output[pos], &temp, sizeof(ptf_collection_on_disk_node));
          pos += sizeof(ptf_collection_on_disk_node);

          write_queue.push(i + 2);
          ++queue_location;
          index_locations[i + 2] = queue_location;
        }
      }
    }



    assert(output.size() == pos);

    *o_data = &output[0];
    *o_size = output.size();
  }
  catch (...) {
    return peTooBigForMachine;
  }

  return peSuccess;
}

// Assumes that parent/child relationship is correct, and so does not check exact positions
ptf_32 get_child_idx(ptf_32 child_left, ptf_32 child_bottom, ptf_32 child_yon, ptf_32 parent_left, ptf_32 parent_bottom, ptf_32 parent_yon)
{
  ptf_32 child_idx = 0;

  if(parent_left != child_left) {
    child_idx |= ptf_child_bit_right;
  }
  if(parent_bottom != child_bottom) {
    child_idx |= ptf_child_bit_top;
  }
  if(parent_yon != child_yon) {
    child_idx |= ptf_child_bit_hither;
  }

  return child_idx;
}

ptf_err ptf_collection_create(ptf_collection_handle *o_coll, ptf_char const * name, ptf_type type)
{
  try {
    size_t len = 0;
    if( name ) {
      len = strlen((char const *)name);
    }
    if (len >= sizeof((*o_coll)->header.name)) {
      return peInvalidArguments;
    }
    *o_coll = new ptf_collection_t(type);
    if( name ) {
      memcpy((*o_coll)->header.name, name, len+1);
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_collection_destroy(ptf_collection_handle coll)
{
  try {
    delete coll;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_collection_get_header(ptf_collection_handle coll, ptf_collection_header * hdr, 
    ptf_record_ref const **o_related, ptf_32 *o_related_count)
{
  try {
    coll->header.related = (ptf_32)coll->related_.size();
    *hdr = coll->header;
    if (o_related) {
      if (coll->related_.size()) {
        *o_related = &coll->related_[0];
      }
      else {
        *o_related = 0;
      }
    }
    if (o_related_count) {
      *o_related_count = hdr->related;
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_collection_set_related(ptf_collection_handle coll,
    ptf_record_ref const *related, ptf_32 count)
{
  try {
    coll->related_.resize(related ? count : 0);
    if (count) {
      memcpy(&coll->related_[0], related, count * sizeof(ptf_record_ref));
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_collection_node_add(ptf_collection_handle coll, ptf_32 parent_id, float f_size, ptf_remote_ref mesh, ptf_8 child_idx, ptf_32 *o_id)
{
  if ((parent_id < 1) || (parent_id > coll->header.num_nodes)) {
    return peInvalidArguments;
  }
  if (child_idx > 7) {
    return peInvalidArguments;
  }

  ptf_collection_on_disk_node new_node;

  new_node.cld_and_parent = ((child_idx&7) << 29) | (parent_id & ((1UL<<29)-1));
  new_node.feature_size = f_size;
  new_node.mesh = mesh;

  // No duplicates allowed
  for (ptf_32 i = 0; i < coll->child_nodes.size(); ++i) {
    if(new_node.cld_and_parent == coll->child_nodes[i].cld_and_parent) {
      return peInvalidArguments;
    }
  }

  coll->records.clear();

  coll->child_nodes.push_back(new_node);
  ++coll->header.num_nodes;

  *o_id = coll->header.num_nodes;

  return peSuccess;
}

ptf_err ptf_collection_set_root(ptf_collection_handle coll, float f_size, ptf_remote_ref mesh, ptf_ocube bounds, ptf_32 *o_id)
{
  if (f_size <= 0.0) {
    return peInvalidArguments;
  }

  // Must not have a root already
  if (coll->root_set) {
    return peInvalidArguments;
  }

  // Make sure it's cubical
  ptf_32 side = bounds.x_max - bounds.x_min;
  if ((bounds.y_max - bounds.y_min) != side) {
    return peInvalidArguments;
  }
  if ((bounds.z_max - bounds.z_min) != side) {
    return peInvalidArguments;
  }

  // Side length must be a power of two
  if ((side & (side-1)) != 0) {
    return peInvalidArguments;
  }

  //  Allow each component to be off by one half
  ptf_32 sDiff = (side >> 1) - 1;
  if ((sDiff & bounds.x_min) || (sDiff & bounds.x_max)) {
    return peInvalidArguments;
  }
  if ((sDiff & bounds.y_min) || (sDiff & bounds.y_max)) {
    return peInvalidArguments;
  }
  if ((sDiff & bounds.z_min) || (sDiff & bounds.z_max)) {
    return peInvalidArguments;
  }

  ptf_32 level;
  ptf_err err = ptf_box_side_to_level( side, &level );
  if (err) {
    return err;
  }

  // Set the header
  coll->header.octree_level = level;
  coll->overall_bounds.x_min = coll->header.left = bounds.x_min;
  coll->overall_bounds.x_max = coll->overall_bounds.x_min + side;
  coll->overall_bounds.y_min = coll->header.bottom = bounds.y_min;
  coll->overall_bounds.y_max = coll->overall_bounds.y_min + side;
  coll->overall_bounds.z_min = coll->header.yon = bounds.z_min;
  coll->overall_bounds.z_max = coll->overall_bounds.z_min + side;

  coll->root_node.left = bounds.x_min;
  coll->root_node.bottom = bounds.y_min;
  coll->root_node.yon = bounds.z_min;
  coll->root_node.disk_node.cld_and_parent = 0;
  coll->root_node.disk_node.feature_size = f_size;
  coll->root_node.disk_node.mesh = mesh;

  coll->records.clear();

  coll->root_set = true;
  ++coll->header.num_nodes;

  *o_id = 1;

  return peSuccess;
}

ptf_err ptf_collection_get_records(ptf_collection_handle coll, ptf_crec_t **o_recs, ptf_32 *o_num_recs)
{
  try {
    coll->records.resize(coll->header.num_nodes);
    if(coll->records.size() != coll->header.num_nodes) {
      return peOutOfMemory;
    }

    ptf_ocube top_level_box;
    ptf_32 box_side;
    ptf_level_to_box_side( coll->header.octree_level, &box_side );

    top_level_box.x_min = coll->header.left;
    top_level_box.x_max = coll->header.left + box_side;
    top_level_box.y_min = coll->header.bottom;
    top_level_box.y_max = coll->header.bottom + box_side;
    top_level_box.z_min = coll->header.yon;
    top_level_box.z_max = coll->header.yon + box_side;

    // Read root
    ptf_crec_t new_rec;
    new_rec.feature_size = coll->root_node.disk_node.feature_size;
    new_rec.mesh = coll->root_node.disk_node.mesh;
    new_rec.child_idx = 0; // Root's child and parent are always zero
    new_rec.yon = top_level_box.z_min;
    new_rec.bottom = top_level_box.y_min;
    new_rec.left = top_level_box.x_min;
    new_rec.children = new_rec.parent = NULL;
    new_rec.num_children = 0;
    new_rec.level = coll->header.octree_level;
    coll->records[0] = new_rec;

    // Read children nodes
    ptf_16 num_child_nodes = coll->header.num_nodes - 1;
    for (ptf_16 i = 0; i < num_child_nodes; ++i) {
      ptf_crec_t new_rec;
      ptf_32 parent_idx = (coll->child_nodes[i].cld_and_parent & ((1UL<<29)-1)) - 1;
      ptf_16 child_idx = coll->child_nodes[i].cld_and_parent >> 29;

      assert(parent_idx < coll->records.size());
      new_rec.child_idx = child_idx;
      new_rec.parent = &(coll->records[parent_idx]);
      new_rec.level = coll->records[parent_idx].level + 1;
      new_rec.left = coll->records[parent_idx].left;
      new_rec.bottom = coll->records[parent_idx].bottom;
      new_rec.yon = coll->records[parent_idx].yon;
      new_rec.feature_size = coll->child_nodes[i].feature_size;
      new_rec.mesh = coll->child_nodes[i].mesh;

      if(child_idx & ptf_child_bit_hither) {
        new_rec.yon += box_side >> (new_rec.level - coll->header.octree_level);
      }
      if(child_idx & ptf_child_bit_top) {
        new_rec.bottom += box_side >> (new_rec.level - coll->header.octree_level);
      }
      if(child_idx & ptf_child_bit_right) {
        new_rec.left += box_side >> (new_rec.level - coll->header.octree_level);
      }

      new_rec.children = NULL;
      new_rec.num_children = 0;
      if(new_rec.parent->children == NULL) {
        new_rec.parent->children = &(coll->records[i+1]);
      }
      ++new_rec.parent->num_children;
      coll->records[i+1] = new_rec;
    }
    assert( coll->records.size() == coll->header.num_nodes );
    *o_recs = &coll->records[0];
    *o_num_recs = coll->header.num_nodes;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_collection_read_from_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptCollectionData;
    ref.index = fref.index;

    //  read header and verify values
    ptf_collection_header hdr;
    ptf_err err = ptf_file_read_record(file, &ref, 0, &hdr, sizeof(hdr));
    if (err == peSuccess) {
      coll->related_.resize(hdr.related);
      if (hdr.related > 0) {
        err = ptf_file_read_record(file, &ref, hdr.related_offset, &coll->related_[0], 
            hdr.related * sizeof(ptf_record_ref));
      }
    }
    ptf_32 cnt = 0;
    ptf_64 datapos = hdr.data_offset;
    if (err == peSuccess) {
      if (hdr.data_offset < sizeof(hdr)) {
        return peInvalidFile;
      }
      if ((hdr.data_offset & 7) != 0) {
        return peInvalidFile;
      }
      if (hdr.num_nodes == 0 || hdr.num_nodes > 0x7fff) {
        return peInvalidFile;
      }
      if (hdr.octree_level > 30) {
        return peInvalidFile;
      }

      std::vector<ptf_collection_on_disk_node> temp_nodes;
      temp_nodes.resize(hdr.num_nodes);
      if(temp_nodes.size() != hdr.num_nodes) {
        return peOutOfMemory;
      }
      err = ptf_file_read_record(file, &ref, hdr.data_offset, &temp_nodes[0], sizeof(ptf_collection_on_disk_node)*hdr.num_nodes);
      if (err == peSuccess) {

        ptf_32 i;
        ptf_32 side;
        ptf_level_to_box_side(hdr.octree_level+1, &side); // Getting side length of a root node, which is one down from the header's level
        for(i = 0; (i < hdr.num_nodes) && (0 == (temp_nodes[i].cld_and_parent & ((1UL<<29)-1))); ++i) {
          ptf_internal_disk_node new_root;
          new_root.disk_node = temp_nodes[i];
          ptf_32 child_idx = new_root.disk_node.cld_and_parent >> 29;
          if(child_idx & ptf_child_bit_hither) {
            new_root.yon = hdr.yon + side;
          }
          else { // Yon
            new_root.yon = hdr.yon;
          }
          if(child_idx & ptf_child_bit_top) {
            new_root.bottom = hdr.bottom + side;
          }
          else { // Bottom
            new_root.bottom = hdr.bottom;
          }
          if(child_idx & ptf_child_bit_right) {
            new_root.left = hdr.left + side;
          }
          else { // Left
            new_root.left = hdr.left;
          }
          coll->root_node = new_root;
          coll->root_set = true;
        }

        assert(coll->root_set);
        coll->child_nodes.resize(hdr.num_nodes - 1);
        if(coll->child_nodes.size() != (hdr.num_nodes - 1)) {
          return peOutOfMemory;
        }
        if (hdr.num_nodes > 1) {
          memcpy( &coll->child_nodes[0], &temp_nodes[1], sizeof(ptf_collection_on_disk_node)*(hdr.num_nodes - 1));
        }
        coll->header = hdr;
      }
    }
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_collection_update_to_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref const &ref)
{
  void * data = 0;
  size_t size = 0;
  ptf_err err;
  try {
    err = coll->marshal_data(&data, &size);
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptCollectionData;
    rref.index = ref.index;
    if (err == peSuccess) {
      err = ptf_file_replace_record(file, &rref);
    }
    if (err == peSuccess) {
      err = ptf_file_write_record(file, data, (ptf_32)size);
    }
    if (err == peSuccess) {
      err = ptf_file_commit_record(file, &rref);
    }
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_collection_append_to_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref * o_ref)
{
  void * data = 0;
  size_t size = 0;
  ptf_err err;
  try {
    err = coll->marshal_data(&data, &size);
    if (err == peSuccess) {
      ptf_record_ref rref;
      err = ptf_file_new_complete_record(file, ptCollectionData, data, (ptf_32)size, &rref);
      o_ref->index = rref.index;
    }
    return err;
  }
  catch (...) {
    return peException;
  }
}
