#include "ptf.h"
#include "ptf_lod.h"
#include "ptf_str.h"
#include "ptf_date.h"
#include "ptf_util.h"

using namespace impl;

#include <assert.h>
#include <string.h>

#include <vector>
#include <algorithm>

class ptf_lod_t {
  public:
                              ptf_lod_t();
                              ~ptf_lod_t();

    std::vector<std::pair<ptf_lod_semantic_header, std::vector<char> > >
                              semantics_;

    ptf_err                   set_semantic_instance(ptf_lod_semantic_header const &sem, 
                                  void const *data, ptf_32 size, ptf_16 cnt = 1);
    ptf_err                   write_to_current_record(ptf_file_handle file);
};

ptf_lod_t::ptf_lod_t()
{
  semantics_.reserve(3);     //  name, lod table, collision mesh without re-allocating vector-of-vectors
}

ptf_lod_t::~ptf_lod_t()
{
}

ptf_err ptf_lod_t::set_semantic_instance(ptf_lod_semantic_header const &sem, 
    void const *data, ptf_32 size, ptf_16 cnt)
{
  if (sem._reserved != 0) {
    return peInvalidArguments;
  }
  if (size & 7) {
    return peInvalidArguments;
  }
  //  size may include padding
  if (sem.item_size && sem.item_size * (ptf_32)cnt > size) {
    return peInvalidArguments;
  }
  for (size_t i = 0; i < semantics_.size(); ++i) {
    if (semantics_[i].first.item_semantic == sem.item_semantic) {
      //  this is it
      if (sem.item_size != semantics_[i].first.item_size) {
        return peInvalidArguments;
      }
      if (sem.item_size && (sem.item_size != size)) {
        return peInvalidArguments;
      }
      semantics_[i].second.clear();
      semantics_[i].second.insert(semantics_[i].second.end(), (char const *)data, (char const *)data + size);
      semantics_[i].first.num_items = cnt;
      semantics_[i].first.semantic_size = size;
      return peSuccess;
    }
  }
  //  I need to add this semantic
  semantics_.push_back(std::pair<ptf_lod_semantic_header, std::vector<char> >(sem, std::vector<char>()));
  semantics_.back().second.insert(semantics_.back().second.end(), (char const *)data, (char const *)data + size);
  ptf_lod_semantic_header &shed = semantics_.back().first;
  shed.num_items = cnt;
  shed.semantic_size = size;
  return peSuccess;
}

ptf_err ptf_lod_t::write_to_current_record(ptf_file_handle file)
{
  ptf_lod_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  hdr.data_offset = sizeof(hdr);
  hdr.num_semantics = (ptf_16)semantics_.size();
  ptf_err err = ptf_file_write_record(file, &hdr, (ptf_32)sizeof(hdr));
  for (size_t i = 0; i < semantics_.size() && err == peSuccess; ++i) {
    ptf_lod_semantic_header shdr = semantics_[i].first;
    assert(shdr.semantic_size == semantics_[i].second.size());
    err = ptf_file_write_record(file, &shdr, (ptf_32)sizeof(shdr));
    if (err == peSuccess) {
      ptf_file_write_record(file, &semantics_[i].second[0], shdr.semantic_size);
    }
  }
  return err;
}


ptf_err ptf_lod_create(ptf_lod_handle *o_lod, ptf_char const * name)
{
  ptf_err err = peSuccess;
  try {
    *o_lod = new ptf_lod_t();
    if (name != NULL) {
      ptf_lod_semantic_header pls;
      memset(&pls, 0, sizeof(pls));
      pls.item_semantic = plsGroupName;
      pls.item_size = 0;
      pls.num_items = 1;
      ptf_32 siz = (ptf_32)strlen((char const *)name)+1;
      pls.semantic_size = (siz + 7) & -8;
      std::vector<char> ch;
      ch.resize(pls.semantic_size);
      memset(&ch[0], 0, pls.semantic_size);
      memcpy(&ch[0], name, siz);
      err = (*o_lod)->set_semantic_instance(pls, &ch[0], pls.semantic_size);
      if (err != peSuccess) {
        delete *o_lod;
        *o_lod = 0;
      }
    }
  }
  catch (...) {
    return peException;
  }
  return err;
}

ptf_err ptf_lod_destroy(ptf_lod_handle lod)
{
  try {
    delete lod;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_read_from_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    lod->semantics_.clear();

    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptLodData;
    ref.index = fref.index;

    //  read header and verify values
    ptf_lod_header hdr;
    ptf_err err = ptf_file_read_record(file, &ref, 0, &hdr, sizeof(hdr));
    ptf_32 cnt = 0;
    ptf_32 datapos = hdr.data_offset;
    if (err == peSuccess) {
      if (hdr.data_offset < sizeof(hdr)) {
        return peInvalidFile;
      }
      if ((hdr.data_offset & 7) != 0) {
        return peInvalidFile;
      }
      if (hdr.num_semantics == 0 || hdr.num_semantics > 0x7fff) {
        return peInvalidFile;
      }
    }
    for (ptf_32 i = 0; i < hdr.num_semantics && err == peSuccess; ++i) {
      ptf_lod_semantic_header shdr;
      err = ptf_file_read_record(file, &ref, datapos, &shdr, sizeof(shdr));
      if (err == peSuccess) {
        datapos += sizeof(shdr);
        if (shdr.item_size * (ptf_32) shdr.num_items > shdr.semantic_size) {
          return peInvalidFile;
        }
        if (shdr.semantic_size & 7) {
          return peInvalidFile;
        }
        lod->semantics_.push_back(std::pair<ptf_lod_semantic_header, std::vector<char> >(shdr, std::vector<char>()));
        lod->semantics_.back().second.resize(shdr.semantic_size);
        err = ptf_file_read_record(file, &ref, datapos, &lod->semantics_.back().second[0], shdr.semantic_size);
        datapos += shdr.semantic_size;
      }
    }
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_update_to_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref const &ref)
{
  try {
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptLodData;
    rref.index = ref.index;
    ptf_err err = ptf_file_replace_record(file, &rref);
    if (err != peSuccess) return err;
    err = lod->write_to_current_record(file);
    if (err != peSuccess) return err;
    err = ptf_file_commit_record(file, &rref);
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_append_to_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref *o_ref)
{
  try {
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    ptf_err err = ptf_file_new_record(file, ptLodData);
    if (err != peSuccess) return err;
    err = lod->write_to_current_record(file);
    if (err != peSuccess) return err;
    err = ptf_file_commit_record(file, &rref);
    o_ref->index = rref.index;
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_count_semantics(ptf_lod_handle lod,
    ptf_32 *o_count)
{
  try {
    *o_count = (ptf_32)lod->semantics_.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_get_semantic_at(ptf_lod_handle lod,
    ptf_32 index, ptf_32 *o_semantic, ptf_16 *o_count, ptf_16 *o_itemsize, 
    ptf_32 *o_datasize)
{
  try {
    if (index >= lod->semantics_.size()) {
      return peInvalidArguments;
    }
    *o_semantic = lod->semantics_[index].first.item_semantic;
    *o_count = lod->semantics_[index].first.num_items;
    *o_itemsize = lod->semantics_[index].first.item_size;
    *o_datasize = (ptf_32)lod->semantics_[index].second.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_get_semantic(ptf_lod_handle lod,
    ptf_32 semantic, ptf_32 *o_index, ptf_16 *o_count, ptf_16 *o_itemsize, 
    ptf_32 *o_datasize)
{
  try {
    ptf_32 index, end = (ptf_32)lod->semantics_.size();
    for (index = 0; index < end; ++index) {
      if (lod->semantics_[index].first.item_semantic == semantic) {
        break;
      }
    }
    if (index >= lod->semantics_.size()) {
      return peInvalidArguments;
    }
    *o_index = index;
    *o_count = lod->semantics_[index].first.num_items;
    *o_itemsize = lod->semantics_[index].first.item_size;
    *o_datasize = (ptf_32)lod->semantics_[index].second.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_get_data(ptf_lod_handle lod,
    ptf_32 index, void *o_data, ptf_32 datasize)
{
  try {
    if (index >= lod->semantics_.size()) {
      return peInvalidArguments;
    }
    if (datasize > lod->semantics_[index].second.size()) {
      datasize = (ptf_32)lod->semantics_[index].second.size();
    }
    if (datasize > 0) { //  avoid taking the address of a non-existent vector element
      memcpy(o_data, &lod->semantics_[index].second[0], datasize);
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_set_semantic_blob(ptf_lod_handle lod,
    ptf_32 semantic, ptf_32 data_size, void const *src)
{
  try {
    ptf_lod_semantic_header sem;
    memset(&sem, 0, sizeof(sem));
    sem.item_semantic = semantic;
    sem.item_size = 0;
    sem.num_items = 1;
    ptf_32 padded_size = (data_size + 7) & -8;
    sem.semantic_size = padded_size;
    std::vector<char> data;
    data.resize(padded_size);
    memset(&data[0], 0, padded_size);
    memcpy(&data[0], src, data_size);
    return lod->set_semantic_instance(sem, &data[0], padded_size);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_set_semantic_items(ptf_lod_handle lod,
    ptf_32 semantic, ptf_16 item_size, ptf_16 num_items, void const *src)
{
  try {
    ptf_lod_semantic_header sem;
    memset(&sem, 0, sizeof(sem));
    sem.item_semantic = semantic;
    sem.item_size = item_size;
    sem.num_items = num_items;
    ptf_32 data_size = (ptf_32)item_size * (ptf_32)num_items;
    ptf_32 padded_size = (data_size + 7) & -8;
    sem.semantic_size = padded_size;
    std::vector<char> data;
    data.resize(padded_size);
    memset(&data[0], 0, padded_size);
    memcpy(&data[0], src, data_size);
    return lod->set_semantic_instance(sem, &data[0], padded_size, num_items);
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_lod_set_table(ptf_lod_handle lod,
    ptf_lod_table_entry const *table, ptf_16 count)
{
  return ptf_lod_set_semantic_items(lod, plsLodTable, sizeof(*table), count, table);
}

ptf_err ptf_lod_get_table(ptf_lod_handle lod,
    ptf_lod_table_entry *o_table, ptf_16 tablecount, ptf_16 *o_count)
{
  try {
    ptf_32 index = 0;
    ptf_16 size = 0;
    ptf_32 dsize = 0;
    ptf_err err = ptf_lod_get_semantic(lod, plsLodTable, &index, o_count, &size, &dsize);
    if (err != peSuccess) return err;
    err = ptf_lod_get_data(lod, index, o_table, tablecount * (ptf_32)sizeof(*o_table));
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_lod_set_collision(ptf_lod_handle lod, 
    ptf_record_ref coll)
{
  return ptf_lod_set_semantic_blob(lod, plsOdeTrimeshCollision, sizeof(ptf_record_ref), &coll);
}

ptf_err ptf_lod_get_collision(ptf_lod_handle lod, 
    ptf_record_ref * o_coll)
{
  ptf_32 index;
  ptf_16 count;
  ptf_16 itemsize;
  ptf_32 datasize;

  try {
    ptf_err err = ptf_lod_get_semantic(lod, plsOdeTrimeshCollision, &index, &count, &itemsize, &datasize);
    if (err != peSuccess) return err;
    if (datasize != sizeof(ptf_record_ref)) {
      return peInvalidArguments;
    }
    err = ptf_lod_get_data(lod, index, o_coll, datasize);
    return err;
  }
  catch(...) {
    return peException;
  }
}
