
#if !defined(ptf_octree_h)
#include "ptf_octree.h"
#endif

#include <assert.h>
#include <math.h>
#include <list>
#include <vector>

#include "mem_util.h"
//  Can't have more than a mega blocks per file
#define MAX_BLOCKS_IN_FILE 0x100000UL
//  Blocks can't be bigger than a megabyte
#define MAX_BLOCK_SIZE 0x100000UL

//  Conversion from meters to octreecoords, so the smallest 
//  representable octree cube is 1/16th of a meter. That can, 
//  in turn, refer to a collection that's split another 12 
//  times.
#define GEOPOS_RATIO 16



//  This is a secret interface between octree and file, because of 
//  the single-instance nature of the octree.
extern void *ptf_file_get_octree_handle(ptf_file_handle file);
extern void ptf_file_set_octree_handle(ptf_file_handle file, void *oh);



static ptf_32 magnitude_diff(ptf_32 a, ptf_32 b)
{
  if (a == b) {
    return 0;
  }
  ptf_32 diff = 1;
  while (!(a&1) && !(b&1)) {
    //  One of the guys will have a bit set before we reach the end
    assert(diff < 0x80000000UL);
    diff <<= 1;
    a >>= 1;
    b >>= 1;
  }
  return diff;
}

//  Todo: re-implement in terms of ptf_octree_get_box_info()
static ptf_16 cube_level(ptf_ocube const & cube, bool offset)
{
  ptf_32 diff = cube.x_max - cube.x_min;
  assert(!(diff & (diff - 1)));
  assert(!(cube.x_max & ((diff - 1) >> (offset ? 1 : 0))));
  assert(!(cube.x_min & ((diff - 1) >> (offset ? 1 : 0))));
  ptf_32 mask = 1;
  ptf_16 ret = 31;
  while (!(mask & diff) && ret > 0) {
    mask <<= 1;
    ret--;
  }
  return ret;
}

//  Todo: re-implement in terms of ptf_octree_get_box_info()
static bool validate_cube_level(ptf_ocube const & cube, ptf_16 level, bool offset)
{
  ptf_32 mask = (1UL << (31-level))-1;
  ptf_32 half = mask;
  if (offset) {
    mask >>= 1;
  }
  if ((cube.x_min & mask) || (cube.x_max & mask) 
      || (cube.y_min & mask) || (cube.y_max & mask)
      || (cube.z_min & mask) || (cube.z_max & mask)) {
    return false;
  }
  if (cube.x_max != cube.x_min + half + 1) {
    return false;
  }
  if (cube.y_max != cube.y_min + half + 1) {
    return false;
  }
  if (cube.z_max != cube.z_min + half + 1) {
    return false;
  }
  return true;
}


namespace {

#if !defined(NDEBUG)
  class tester {
    public:
      tester() {
        ptf_ocube c = { 0, 0x80000000, 0, 0x80000000, 0, 0x80000000 };
        assert(cube_level(c, false) == 0);
        ptf_ocube d = { 0, 0x40000000, 0, 0x40000000, 0, 0x40000000 };
        assert(cube_level(d, false) == 1);
        d.x_min += 0x20000000;
        d.x_max += 0x20000000;
        assert(cube_level(d, true) == 1);
      }
  };

  tester t;
#endif


  struct octree_block {
    octree_block(ptf_octree_t * oct)
      {
        dirty_ = false;
        new_ = true;
        octree_ = oct;

        //  the first node is the root of this branch of the tree
        ptf_octree_node on;
        memset(&on, 0, sizeof(on));
        ptf_fixed_ref fr;
        memset(&fr, 0, sizeof(fr));
        add_node(on, fr);
      }
    ptf_octree_block_header           header_;
    std::vector<ptf_octree_node>      nodes_;
    std::vector<ptf_fixed_ref>        content_;
    std::vector<ptf_remote_ref>       links_;
    bool                              dirty_;
    bool                              new_;
    ptf_octree_t *                    octree_;
    //  0 is an invalid value (it's the root)
    ptf_16                            add_node(ptf_octree_node const & on, ptf_fixed_ref const & fr);
    ptf_err                           add_link(ptf_remote_ref const & rr, ptf_16 *o_link);
  };

  struct coord3 {
    ptf_32 x, y, z;
    coord3() { x = 0; y = 0; z = 0; }
    //  Note: The function coord3::set() is the normative function 
    //  for converting a geopos into an octree coordinate. Also look 
    //  at the function ptf_octree_box_get_coords().
    ptf_err set(ptf_geopos const & gp) {
      if (gp.x < PTF_MIN_GEOPOS || gp.x >= PTF_MAX_GEOPOS) {
        return peInvalidArguments;
      }
      if (gp.y < PTF_MIN_GEOPOS || gp.y >= PTF_MAX_GEOPOS) {
        return peInvalidArguments;
      }
      if (gp.z < PTF_MIN_GEOPOS || gp.z >= PTF_MAX_GEOPOS) {
        return peInvalidArguments;
      }
      x = (ptf_32)floor((gp.x - PTF_MIN_GEOPOS) * GEOPOS_RATIO);
      y = (ptf_32)floor((gp.y - PTF_MIN_GEOPOS) * GEOPOS_RATIO);
      z = (ptf_32)floor((gp.z - PTF_MIN_GEOPOS) * GEOPOS_RATIO);
      return peSuccess;
    }
  };
}


struct ptf_octree_t {
  ptf_octree_t(ptf_file_handle file) : file_(file)
    {
      sig_ = 'oct ';
      refCount_ = 1;
      ptf_file_set_octree_handle(file, this);
    }
  ~ptf_octree_t()
    {
      if (file_ != 0) {
        assert(ptf_file_get_octree_handle(file_) == this);
        ptf_file_set_octree_handle(file_, 0);
      }
      for (std::vector<octree_block *>::iterator ptr = blocks_.begin(), 
          end = blocks_.end(); ptr != end; ++ptr) {
        delete *ptr;
      }
      sig_ = 0;
    }

  bool validate() { return sig_ == 'oct '; }

  ptf_err open(ptf_ocube * range, ptf_16 max_depth, ptf_16 max_nodes, bool offset);
  ptf_err load_blocks(bool offset);
  //  Create a new block, populated with a single
  //  (root) node, and zero links.
  ptf_err new_block(ptf_32 * o_ix, ptf_16 level);
  ptf_err flush();

  ptf_32                        sig_;
  ptf_32                        refCount_;
  ptf_file_handle               file_;
  ptf_file_header               header_;
  ptf_16                        maxDepth_;
  ptf_16                        maxNodes_;
  ptf_16                        rootLevel_;
  std::vector<octree_block *>   blocks_;
};




ptf_16 octree_block::add_node(ptf_octree_node const & on, ptf_fixed_ref const & fr)
{
  if (nodes_.size() == octree_->maxNodes_) {
    return 0;
  }
  nodes_.push_back(on);
  content_.push_back(fr);
  dirty_ = true;
  return (ptf_16)nodes_.size()-1;
}

ptf_err octree_block::add_link(ptf_remote_ref const & rr, ptf_16 *o_link)
{
  if (links_.size() >= 0x8000) {
    return peTooBigForMachine;
  }
  links_.push_back(rr);
  dirty_ = true;
  *o_link = (ptf_16)links_.size()-1;
  return peSuccess;
}

static ptf_err ptf_onode_split_internal(ptf_octree_handle oh, octree_block *ob, ptf_onode_ref const &parent, 
    coord3 const &c3, ptf_onode_ref *o_ref, bool forceNew = false)
{
  *o_ref = parent;
  ptf_32 diff = (parent.extent.x_max - parent.extent.x_min) >> 1;
  ptf_32 x = (c3.x >= (parent.extent.x_min + diff) ? 1 : 0);
  ptf_32 y = (c3.y >= (parent.extent.y_min + diff) ? 1 : 0);
  ptf_32 z = (c3.z >= (parent.extent.z_min + diff) ? 1 : 0);
  o_ref->extent.x_min += x * diff;
  o_ref->extent.x_max -= (1-x) * diff;
  o_ref->extent.y_min += y * diff;
  o_ref->extent.y_max -= (1-y) * diff;
  o_ref->extent.z_min += z * diff;
  o_ref->extent.z_max -= (1-z) * diff;
  o_ref->node_level++;
  ptf_16 next = ob->nodes_[parent.node_index].child[x][y][z];
  if (next == 0) {
    //  create a split
    if (!forceNew && ob->nodes_.size() < oh->maxNodes_ && (o_ref->node_level - ob->header_.level) < oh->maxDepth_) {
      //  use the current block
      ptf_octree_node nod;
      memset(&nod, 0, sizeof(nod));
      ptf_fixed_ref fr;
      memset(&fr, 0, sizeof(fr));
      o_ref->node_index = ob->add_node(nod, fr);
      ob->nodes_[parent.node_index].child[x][y][z] = o_ref->node_index;
    }
    else {
      //  must allocate a new block
      octree_block * oblock = ob;
      assert(o_ref->node_level > 0);
      ptf_err ret = oh->new_block(&o_ref->octree_index, o_ref->node_level);
      if (ret != peSuccess) {
        return ret;
      }
      ob = oh->blocks_[o_ref->octree_index];
      o_ref->node_index = 0;
      ptf_remote_ref rref;
      rref.file = 0;
      rref.index = o_ref->octree_index;
      ptf_16 link = 0;
      if ((ret = oblock->add_link(rref, &link)) != peSuccess) {
        return ret;
      }
      oblock->nodes_[parent.node_index].child[x][y][z] = link | 0x8000;
      oblock->dirty_ = true;
    }
    ob->dirty_ = true;
  }
  else if (next & 0x8000) {
    //  follow a block link
    if ((ptf_32)(next & 0x7fff) >= ob->links_.size()) {
      return peInvalidFile;
    }
    ptf_remote_ref rref = ob->links_[next & 0x7fff];
    o_ref->node_index = 0;
    o_ref->octree_index = rref.index;
    if (rref.file != 0) {
      return peInvalidArguments;
    }
    if (rref.index >= oh->blocks_.size()) {
      return peInvalidFile;
    }
    ob = oh->blocks_[rref.index];
  }
  else {
    //  follow an internal link
    o_ref->node_index = next;
    if (o_ref->node_index >= ob->nodes_.size()) {
      return peInvalidFile;
    }
  }
  o_ref->content_index = ob->content_[o_ref->node_index].index;
  return peSuccess;
}

ptf_err ptf_octree_t::open(ptf_ocube * range, ptf_16 max_depth, ptf_16 max_nodes, bool offset)
{
  ptf_32 rootLevel;

  rootLevel = cube_level(*range, offset);
  if (!validate_cube_level(*range, rootLevel, offset)) {
    return peInvalidArguments;
  }

  maxDepth_ = max_depth;
  maxNodes_ = max_nodes;

  ptf_err ret = ptf_file_get_header(file_, &header_);
  if (ret != peSuccess) {
    return ret;
  }

  ptf_record_ref ref;
  memset(&ref, 0, sizeof(ref));
  ref.type = ptOctreeNode;
  ref.index = 1;
  ptf_64 r_size = 0;
  ret = ptf_file_get_record_size(file_, &ref, &r_size);
  switch (ret) {
    case peSuccess:           //  The record is there, load it
      ret = load_blocks(offset);
      assert(blocks_[0] == 0);
      assert(blocks_[1] != 0);
      *range = header_.range;
      break;
    case peTypeNotFound:      //  I need to create the record
      assert(blocks_.size() == 0);
      blocks_.push_back(0);
      (void)new_block(0, 0);
      assert(blocks_[0] == 0);
      assert(blocks_[1] != 0);
      if (offset) {
        //  create eight child blocks right off the bat!
        assert(blocks_.size() == 2);
        ptf_onode_ref parent, child;
        memset(&parent, 0, sizeof(parent));
        parent.node_index = 0;
        parent.node_level = 0;
        parent.octree_index = 1;
        parent.file_index = 0;
        parent.content_index = 0;
        parent.extent = *range;
        for (int i = 0; i < 8; ++i) {
          coord3 coord;
          coord.x = (i & 1) ? range->x_max : range->x_min;
          coord.y = (i & 2) ? range->y_max : range->y_min;
          coord.z = (i & 4) ? range->z_max : range->z_min;
          ptf_onode_split_internal(this, blocks_[1], parent, coord, &child, true);
          assert(blocks_.size() == i + 3);
        }
      }
      header_.range = *range;
      ret = ptf_file_update_header(file_, &header_);
      break;
    case peRecordNotFound:    //  This file is somehow corrupt!
      ret = peInvalidFile;
    default:                  //  Some other error
      break;
  }
  if (ret == peSuccess) {
    rootLevel_ = rootLevel;
  }

  return ret;
}


ptf_err ptf_octree_t::load_blocks(bool offset)
{
  //  load each block found in the file (don't care about inter-relation 
  //  references right now -- that happens when traversing).
  ptf_32 count = 0;
  ptf_err ret = ptf_file_count_type_records(file_, ptOctreeNode, &count);
  if (ret != peSuccess) {
    return ret;
  }

  assert(blocks_.size() == 0);
  blocks_.push_back(0);
  for (ptf_32 i = 1; i < count; ++i) {

    //  get block reference and size
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptOctreeNode;
    rref.index = i;
    ptf_64 size = 0;
    ret = ptf_file_get_record_size(file_, &rref, &size);
    if (ret != peSuccess) {
      blocks_.push_back(0);
      continue;
    }
    if (size < sizeof(ptf_octree_block_header)) {
      return peInvalidFile;
    }
    if (size > MAX_BLOCK_SIZE) {
      return peInvalidFile;
    }

    //  read block header
    char * data = new char[(size_t)size];
    stack_array_delete<char> del_data(data);
    ret = ptf_file_read_record(file_, &rref, 0, data, size);
    if (ret != peSuccess) {
      return ret;
    }
    octree_block * ob = new octree_block(this);
    stack_delete<octree_block> del_ob(ob);
    memcpy(&ob->header_, data, sizeof(ob->header_));

    //  verify that the block seems reasonable
    if (ob->header_.level > 30 || (ob->header_.level == 0 && i != 1)) {
      //  only the first block can be rooted at level 0
      return peInvalidFile;
    }
    size_t siz1 = ob->header_.num_nodes * sizeof(ptf_octree_node);
    size_t siz2 = ob->header_.num_nodes * sizeof(ptf_fixed_ref);
    size_t siz3 = ob->header_.num_links * sizeof(ptf_remote_ref);
    if (size < sizeof(ob->header_) + siz1 + siz2 + siz3) {
      return peInvalidFile;
    }
    if (size < ob->header_.node_array_pointer + siz1) {
      return peInvalidFile;
    }
    if (size < ob->header_.content_array_pointer + siz2) {
      return peInvalidFile;
    }
    if (size < ob->header_.link_pointer + siz3) {
      return peInvalidFile;
    }

    //  extract the actual data
    ob->new_ = false;
    ob->dirty_ = false;
    ob->nodes_.resize(ob->header_.num_nodes);
    ob->content_.resize(ob->header_.num_nodes);
    ob->links_.resize(ob->header_.num_links);
    if (siz1) {
      memcpy(&ob->nodes_[0], data+ob->header_.node_array_pointer, siz1);
    }
    if (siz2) {
      memcpy(&ob->content_[0], data+ob->header_.content_array_pointer, siz2);
    }
    if (siz3) {
      memcpy(&ob->links_[0], data+ob->header_.link_pointer, siz3);
    }
    del_ob.detach();
    blocks_.push_back(ob);
  }

  assert(blocks_.size() == count);    //  0 is empty
  return peSuccess;
}


ptf_err ptf_octree_t::new_block(ptf_32 * o_ix, ptf_16 level)
{
  if (blocks_.size() >= MAX_BLOCKS_IN_FILE) {
    return peTooBigForMachine;
  }
  octree_block * ob = new octree_block(this);
  memset(&ob->header_, 0, sizeof(ob->header_));
  blocks_.push_back(ob);
  if (o_ix) {
    *o_ix = (ptf_32)blocks_.size()-1;
  }
  ob->dirty_ = true;
  ob->header_.level = level;
  return peSuccess;
}


ptf_err ptf_octree_t::flush()
{
  ptf_err ret = peSuccess;
  assert(!blocks_.size() || blocks_[0] == 0);
  for (std::vector<octree_block *>::iterator ptr = blocks_.begin(), 
      end = blocks_.end(); ptr != end; ++ptr) {
    if (*ptr && (*ptr)->dirty_) {
      //  flush this block to disk
      std::vector<char> data;
      ptf_16 level = (*ptr)->header_.level;
      memset(&(*ptr)->header_, 0, sizeof((*ptr)->header_));
      (*ptr)->header_.num_nodes = (ptf_16)(*ptr)->nodes_.size();
      (*ptr)->header_.num_links = (ptf_16)(*ptr)->links_.size();
      (*ptr)->header_.level = level;
      (*ptr)->header_.node_array_pointer = sizeof((*ptr)->header_);
      size_t siz1 = sizeof(ptf_octree_node) * (*ptr)->nodes_.size();
      (*ptr)->header_.content_array_pointer = (ptf_32)(siz1 + sizeof(ptf_octree_block_header));
      //  round up, just to be nice
      size_t siz2 = (sizeof(ptf_fixed_ref) * (*ptr)->content_.size() + 15) & -16;
      (*ptr)->header_.link_pointer = (ptf_32)(siz2 + siz1 + sizeof(ptf_octree_block_header));
      size_t siz3 = sizeof(ptf_remote_ref) * (*ptr)->links_.size();
      size_t sizTotal = (ptf_32)(siz3 + siz2 + siz1 + sizeof(ptf_octree_block_header));
      assert(sizTotal > 0);

      data.resize(sizTotal);
      memset(&data[0], 0, sizTotal);
      memcpy(&data[0], &(*ptr)->header_, sizeof((*ptr)->header_));
      // if siz=0 for the last header component the subscript will be out of range
      // when taking the address of a vector element. VC9 will barf in debug mode.
      if (siz1 > 0) {
        memcpy(&data[sizeof(ptf_octree_block_header)], &(*ptr)->nodes_[0], siz1);
      }
      if (siz2 > 0) {
        memcpy(&data[sizeof(ptf_octree_block_header) + siz1], 
          &(*ptr)->content_[0], sizeof(ptf_fixed_ref) * (*ptr)->content_.size());
      }
      if (siz3 > 0) {
        memcpy(&data[sizeof(ptf_octree_block_header) + siz1 + siz2], 
          &(*ptr)->links_[0], siz3);
      }
      ptf_record_ref ref;
      if ((*ptr)->new_) {
        ret = ptf_file_new_complete_record(file_, ptOctreeNode, &data[0], (ptf_32)sizTotal, &ref);
        assert(ref.index != 0);
        assert(ref.index == ptr - blocks_.begin());        //  record 0 is empty
      }
      else {
        memset(&ref, 0, sizeof(ref));
        ref.type = ptOctreeNode;
        ref.index = (ptf_32)(ptr - blocks_.begin());       //  0 is empty
        assert(ref.index != 0);
        ret = ptf_file_replace_record(file_, &ref);
        if (ret == peSuccess) {
          ret = ptf_file_write_record(file_, &data[0], (ptf_32)sizTotal);
        }
        if (ret == peSuccess) {
          ret = ptf_file_commit_record(file_, &ref);
        }
      }
      if (ret != peSuccess) {
        return ret;
      }
      (*ptr)->new_ = false;
      (*ptr)->dirty_ = false;
    }
  }
  return ret;
}





ptf_err ptf_octree_create(ptf_file_handle file, ptf_ocube *io_range,
    ptf_octree_handle *o_handle, ptf_16 max_depth, ptf_16 max_nodes)
{
  try {
    //  The range must describe a cube. This means that the 
    //  difference is a power of two.
    if (io_range->x_min >= io_range->x_max 
        || io_range->y_min >= io_range->y_max
        || io_range->z_min >= io_range->z_max) {
      return peInvalidArguments;
    }
    if ((io_range->x_max - io_range->x_min != io_range->y_max - io_range->y_min)
        || (io_range->x_max - io_range->x_min != io_range->z_max - io_range->z_min)) {
      return peInvalidArguments;
    }
    //  Allow each component to be off by one half
    ptf_32 sDiff = ((io_range->x_max - io_range->x_min) >> 1) - 1;
    if ((sDiff & io_range->x_max) || (sDiff & io_range->x_max)) {
      return peInvalidArguments;
    }
    if ((sDiff & io_range->y_max) || (sDiff & io_range->y_max)) {
      return peInvalidArguments;
    }
    if ((sDiff & io_range->z_max) || (sDiff & io_range->z_max)) {
      return peInvalidArguments;
    }

    //  Validate the control parameters.
    if (max_depth == 0) {
      max_depth = 24;
    }
    if (max_nodes == 0) {
      max_nodes = 4096;
    }
    if (max_depth < 2) {
      max_depth = 2;
    }
    if (max_depth > 24) {
      max_depth = 24;
    }
    if (max_nodes < 8) {
      max_nodes = 8;
    }
    if (max_nodes > 4096) {
      max_nodes = 4096;
    }

    *o_handle = (ptf_octree_handle)ptf_file_get_octree_handle(file);
    if (*o_handle) {
      ptf_file_header hdr;
      ptf_err e = ptf_file_get_header(file, &hdr);
      if (e != peSuccess) {
        *o_handle = 0;
        return e;
      }
      (*o_handle)->refCount_++;
      *io_range = hdr.range;
      return peSuccess;
    }
    else {
      //  Attempt to create and open the object instance
      *o_handle = new ptf_octree_t(file);
      //  Figure out whether this cube is offset by half along some 
      //  axis, or not.
      bool offset = false;
      if ((io_range->x_min & (io_range->x_max - io_range->x_min - 1))
          || (io_range->y_min & (io_range->y_max - io_range->y_min - 1))
          || (io_range->z_min & (io_range->z_max - io_range->z_min - 1))) {
        offset = true;
      }
      ptf_err ret = (*o_handle)->open(io_range, max_depth, max_nodes, offset);
      if (ret != peSuccess) {
        --(*o_handle)->refCount_;
        delete *o_handle;
        *o_handle = NULL;
      }
      return ret;
    }
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_close(ptf_octree_handle oh)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    ptf_err ret = oh->flush();
    if (!--oh->refCount_) {
      delete oh;
    }
    return ret;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_resolve_to_level(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_16 target_level, ptf_onode_ref * o_ref)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    //  Make sure the coordinates are within range.
    coord3 c3;
    ptf_err ret = c3.set(pos);
    if (ret != peSuccess) {
      return ret;
    }

    //  Reject the call if the point is outside.
    o_ref->octree_index = 1;
    octree_block * ob = oh->blocks_[1];
    ptf_ocube cube = oh->header_.range;
    if (c3.x < cube.x_min || c3.x >= cube.x_max || 
        c3.y < cube.y_min || c3.y >= cube.y_max ||
        c3.z < cube.z_min || c3.z >= cube.z_max) {
      return peInvalidArguments;
    }

    ptf_16 level = oh->rootLevel_;
    ptf_16 chunkLevel = 0;
    ptf_16 ni = 0;
    ptf_32 diff = cube.x_max - cube.x_min;
    assert(!(diff & (diff - 1)));
    assert((1UL << (31-level)) == diff);

    //  It's time to traverse the octree.
    //  Doing it recursively uses too much stack, and there 
    //  is no reason to.
    while (diff > 0) {
      diff >>= 1;
      ptf_32 x = (c3.x >= (cube.x_min + diff) ? 1 : 0);
      ptf_32 y = (c3.y >= (cube.y_min + diff) ? 1 : 0);
      ptf_32 z = (c3.z >= (cube.z_min + diff) ? 1 : 0);
      ptf_16 next = ob->nodes_[ni].child[x][y][z];
      if ((next == 0) || (level >= target_level)) {
        //  this is as far as we go
        cube.x_max = cube.x_min + diff*2;
        cube.y_max = cube.y_min + diff*2;
        cube.z_max = cube.z_min + diff*2;
        o_ref->content_index = ob->content_[ni].index;
        o_ref->extent = cube;
        o_ref->file_index = 0;
        o_ref->node_index = ni;
        o_ref->node_level = chunkLevel;
        return peSuccess;
      }
      else if (next & 0x8000) {
        //  follow link to a new block
        if ((ptf_32)(next & 0x7fff) >= ob->links_.size()) {
          return peInvalidFile;
        }
        ptf_remote_ref rr = ob->links_[next & 0x7fff];
        if (rr.file == 0) {
          if (rr.index >= oh->blocks_.size()) {
            return peInvalidFile;
          }
          ob = oh->blocks_[rr.index];
          o_ref->octree_index = rr.index;
        }
        else {
          //  TODO: return the appropriate reference here, 
          //  still with the "wrongfile" error code.
          assert(!"implement me!");
          return peWrongFile;
        }
        ni = 0;
        chunkLevel = 0;
      }
      else {
        //  follow link within the same block
        ni = next;
      }
      ++level;
      ++chunkLevel;
      cube.x_min += diff * x;
      cube.y_min += diff * y;
      cube.z_min += diff * z;
    }
    return peInvalidFile;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_resolve(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_onode_ref * o_ref)
{
  // level 30 is the max valid resolution level
  return ptf_octree_resolve_to_level(oh, pos, 30, o_ref);
}


ptf_err ptf_octree_split(ptf_octree_handle oh, ptf_geopos const & pos, 
    ptf_onode_ref const & parent, ptf_onode_ref * o_ref)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    if (parent.file_index != 0) {
      return peWrongFile;
    }
    if (parent.node_level >= 30) {
      return peInvalidArguments;
    }
    coord3 c3;
    if (c3.set(pos) != peSuccess) {
      return peInvalidArguments;
    }
    if (!validate_cube_level(parent.extent, parent.node_level, parent.node_level == oh->rootLevel_)) {
      return peInvalidArguments;
    }
    if (c3.x < parent.extent.x_min || c3.x >= parent.extent.x_max ||
        c3.y < parent.extent.y_min || c3.y >= parent.extent.y_max ||
        c3.z < parent.extent.z_min || c3.z >= parent.extent.z_max) {
      return peInvalidArguments;
    }
    if (parent.octree_index >= oh->blocks_.size()) {
      return peInvalidArguments;
    }
    octree_block * ob = oh->blocks_[parent.octree_index];
    if (parent.node_index >= ob->nodes_.size()) {
      return peInvalidArguments;
    }
    return ptf_onode_split_internal(oh, ob, parent, c3, o_ref);
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_onode_get_children(ptf_octree_handle oh, ptf_onode_ref const & parent,
    ptf_onode_ref * children)
{
  try {
    memset(children, 0, sizeof(ptf_onode_ref)*8);
    if (!oh->validate()) return peInvalidArguments;
    if (parent.file_index != 0) {
      //  You should have followed the file reference!
      return peWrongFile;
    }
    if (parent.node_level > 30) {
      return peInvalidArguments;
    }
    if (!validate_cube_level(parent.extent, parent.node_level, parent.node_level == oh->rootLevel_)) {
      return peInvalidArguments;
    }
    if (parent.octree_index >= oh->blocks_.size()) {
      return peInvalidArguments;
    }
    octree_block * ob = oh->blocks_[parent.octree_index];
    if (parent.node_index >= ob->nodes_.size()) {
      return peInvalidArguments;
    }
    ptf_32 diff = (parent.extent.x_max - parent.extent.x_min) >> 1;
    for (ptf_16 i = 0; i < 8; ++i) {
      octree_block * obtemp = ob;
      ptf_32 x = ((i & 4) ? 1 : 0);
      ptf_32 y = ((i & 2) ? 1 : 0);
      ptf_32 z = ((i & 1) ? 1 : 0);
      size_t numnodes = ob->nodes_.size();
      ptf_16 next = ob->nodes_[parent.node_index].child[x][y][z];
      children[i] = parent;
      children[i].extent.x_min += x * diff;
      children[i].extent.x_max -= (1-x) * diff;
      children[i].extent.y_min += y * diff;
      children[i].extent.y_max -= (1-y) * diff;
      children[i].extent.z_min += z * diff;
      children[i].extent.z_max -= (1-z) * diff;
      children[i].node_level++;
      if (next == 0) {
        //  There is no child -- set the level to 0 as indicator
        children[i].node_level = 0;
        children[i].octree_index = 0;
      }
      else if (next & 0x8000) {
        //  follow a block link
        if ((ptf_32)(next & 0x7fff) >= ob->links_.size()) {
          return peInvalidFile;
        }
        ptf_remote_ref rref = ob->links_[next & 0x7fff];
        children[i].node_index = 0;
        children[i].octree_index = rref.index;
        if (rref.file == 0) {
          if (rref.index >= oh->blocks_.size()) {
            return peInvalidFile;
          }
          obtemp = oh->blocks_[rref.index];
        }
        else {
          //  I need to return a reference saying I'm found in the other file.
          //  That means I have no content here.
          obtemp = 0;
          children[i].file_index = rref.file;
          children[i].content_index = 0;
        }
      }
      else {
        //  follow an internal link
        children[i].node_index = next;
        if (children[i].node_index >= ob->nodes_.size()) {
          return peInvalidFile;
        }
      }
      if (obtemp != 0) {
        children[i].content_index = obtemp->content_[children[i].node_index].index;
      }
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_update_content_index(ptf_octree_handle oh,
    ptf_onode_ref const & ref)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    if (ref.file_index != 0) {
      return peInvalidArguments;
    }
    if (ref.octree_index >= oh->blocks_.size()) {
      return peInvalidArguments;
    }
    octree_block * ob = oh->blocks_[ref.octree_index];
    if (ref.node_index >= ob->nodes_.size()) {
      return peInvalidArguments;
    }
    if (ob->content_[ref.node_index].index != ref.content_index) {
      ob->content_[ref.node_index].index = ref.content_index;
      ob->dirty_ = true;
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_get_root(ptf_octree_handle oh, ptf_onode_ref * o_root)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    if (oh->blocks_.size() < 2 || oh->blocks_[0] || !oh->blocks_[1]) {
      return peInvalidFile;
    }
    memset(o_root, 0, sizeof(*o_root));
    o_root->file_index = 0;
    o_root->octree_index = 1;
    o_root->node_index = 0;
    o_root->node_level = oh->rootLevel_;
    o_root->content_index = oh->blocks_[1]->content_[0].index;
    o_root->extent = oh->header_.range;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_octree_get_box_info(ptf_ocube const &box, ptf_32 *o_level, ptf_32 *o_side, ptf_32 *o_offset)
{
  ptf_32 side = box.x_max - box.x_min;
  if (side < 2) return peInvalidArguments;
  if (side > 0x80000000) return peInvalidArguments;
  if (side & (side - 1)) return peInvalidArguments;
  if (box.y_max - box.y_min != side) return peInvalidArguments;
  if (box.z_max - box.z_min != side) return peInvalidArguments;
  //  I don't have to test min, because I've already verified that the 
  //  distance between max and min is a power of two. Divide by two, 
  //  to make sure that off-by-half is allowed.
  if (box.x_max & (side / 2 - 1)) return peInvalidArguments;
  if (box.y_max & (side / 2 - 1)) return peInvalidArguments;
  if (box.z_max & (side / 2 - 1)) return peInvalidArguments;
  //  Calculate the level based on the side length.
  ptf_err err = ptf_box_side_to_level(side, o_level);
  if (err != peSuccess) return err;
  *o_side = side;
  *o_offset = 0;
  if (box.x_max & (side - 1)) *o_offset |= 1;
  if (box.y_max & (side - 1)) *o_offset |= 2;
  if (box.z_max & (side - 1)) *o_offset |= 4;
  return peSuccess;
}

ptf_err ptf_octree_set_reference(ptf_octree_handle oh, ptf_file_handle ref_file)
{
  try {
    if (!oh->validate()) return peInvalidArguments;
    if (oh->blocks_.size() < 2 || oh->blocks_[0] || !oh->blocks_[1]) {
      return peInvalidFile;
    }
    ptf_ocube extent;
    ptf_file_header hdr;
    ptf_err err;
    ptf_32 side, level, shift;
    ptf_onode_ref refParent, refChildren[8];
    //  0. get or add a reference to the file
    ptf_32 index = 0;
    ptf_file_definition_record def;
    memset(&def, 0, sizeof(def));
    //  extract the path
    err = ptf_file_get_definition_record(ref_file, &def);
    if (err != peSuccess) return err;
    //  look it up (or create it)
    err = ptf_file_make_reference(oh->file_, &def, &index);
    if (err != peSuccess) return err;
    //  1. extract extents from file
    err = ptf_file_get_header(ref_file, &hdr);
    if (err != peSuccess) return err;
    extent = hdr.range;
    //  2. validate extents
    err = ptf_octree_get_box_info(extent, &level, &side, &shift);
    if (err != peSuccess) return err;
    // TODO: 3. verify the file octree

    //  open/create the octree of the referenced file to extract the 
    //  extent and the eight children (if off-by-half).
    ptf_octree_handle refOh = 0;
    err = ptf_octree_create(ref_file, &hdr.range, &refOh, 0, 0);
    if (err != peSuccess) return err;
    err = ptf_octree_get_root(refOh, &refParent);
    memset(refChildren, 0, sizeof(refChildren));
    //  Store the parent in children array if there's only a parent, that 
    //  way the children array contains all the cubes I need to process.
    if (shift != 0) {
      err = ptf_onode_get_children(refOh, refParent, refChildren);
      if (err != peSuccess) return err;
    }
    else {
      refChildren[0] = refParent;
    }
    ptf_onode_ref rememberedTargets[8];
    memset(rememberedTargets, 0, sizeof(rememberedTargets));
    //  4. for each block in the reference (typically 1 or 8)
    for (int i = 0; i < 8; ++i) {
      if (refChildren[i].octree_index == 0) continue;
      ptf_geopos lower, upper;
      err = ptf_octree_box_get_coords(refChildren[i].extent, &lower, &upper);
      if (err != peSuccess) return err;
      //  4.1. split master until at the child cube level
      ptf_onode_ref cur, parent, descendants[8];
      err = ptf_octree_get_root(oh, &cur);
      parent = cur;
      if (err != peSuccess) return err;
      while (true) {
        //  if this cube references a different file, return error
        if (cur.file_index != 0) {
          return peWrongFile;
        }
        //  if this cube is the same as the refchild, I'm done
        if (ptf_ocube_equals_ocube(cur.extent, refChildren[i].extent)) {
          break;
        }
        err = ptf_onode_get_children(oh, cur, descendants);
        if (err != peSuccess) return err;
        int found = -1;
        for (int c = 0; c < 8; ++c) {
          if (descendants[c].node_level != 0 && ptf_ocube_contains_ocube(descendants[c].extent, refChildren[i].extent)) {
            found = c;  //  at most one child can contain the sought cube
            break;
          }
        }
        parent = cur;
        //  if this cube has a child in this quadrant,
        if (found != -1) {
          //  descend this cube, based on the left-lo-yon of the refchild
          cur = descendants[found];
        }
        else {
          //  split this cube, based on the left-lo-yon of the refchild
          err = ptf_octree_split(oh, lower, parent, &cur);
          if (err != peSuccess) return err;
        }
      }
      //  4.2. if master has content, return error
      if (cur.content_index != 0) return peInvalidArguments;
      //  4.3. if master has children at this level, return error
      err = ptf_onode_get_children(oh, cur, descendants);
      if (err != peSuccess) return err;
      for (int q = 0; q < 8; ++q)
        if (descendants[q].node_level != 0)
          return peInvalidArguments;
      //  4.4. remember the mapping
      rememberedTargets[i] = parent;
    }
    //  5. for each remembered mapping
    for (int i = 0; i < 8; ++i) {
      ptf_onode_ref &parent = rememberedTargets[i];
      if (parent.node_level != 0) {
        //  look up the block and node
        assert(oh->blocks_.size() > parent.octree_index);
        octree_block *ob = oh->blocks_[parent.octree_index];
        assert(ob->nodes_.size() > parent.node_index);
        //  create a ref to the root, referenced by this child
        ptf_remote_ref the_ref;
        the_ref.file = index;
        the_ref.index = refChildren[i].octree_index;
        ptf_16 link = 0;
        //  I can't have more than 7fff links in a given block.
        //  That's a problem if I split down to the lowest level of 
        //  nodes, and then make all the nodes have 8 link children. 
        //  However, the enforced limit of 4096 nodes per block takes 
        //  care of this, making it unlikely (but not impossible) I 
        //  will run out of links, as each node can have 8 children.
        //  TODO: when running out of links, re-factor the octree 
        //  data to split it into two records, putting half of the 
        //  links in each.
        err = ob->add_link(the_ref, &link);
        if (err != peSuccess) return err;
        //  Figure out which element of the parent node to set to this reference.
        //  The box of the child should lie along the min or max for each cardinal direction.
        int x = (parent.extent.x_min == refChildren[i].extent.x_min) ? 0 : 1;
        int y = (parent.extent.y_min == refChildren[i].extent.y_min) ? 0 : 1;
        int z = (parent.extent.z_min == refChildren[i].extent.z_min) ? 0 : 1;
        assert( !x || parent.extent.x_max == refChildren[i].extent.x_max );
        assert( !y || parent.extent.y_max == refChildren[i].extent.y_max );
        assert( !z || parent.extent.z_max == refChildren[i].extent.z_max );
#if !defined(NDEBUG)
        //  Note: I don't use this information in the code, I'm just checking that the 
        //  algorithm works as expected. If these fail, there's a bug in the algorithm 
        //  above. Thus, assert() is the right checking mechanism for this instance.
        ptf_32 plevel, pside, poffset;
        err = ptf_octree_get_box_info(parent.extent, &plevel, &pside, &poffset);
        assert( err == peSuccess );
        ptf_32 clevel, cside, coffset;
        err = ptf_octree_get_box_info(refChildren[i].extent, &clevel, &cside, &coffset);
        assert( err == peSuccess );
        assert( clevel == plevel+1 );
        assert( cside == pside >> 1 );
        assert( !poffset );
        assert( !coffset );
#endif
        //  5.1. update the file mapping
        ob->nodes_[parent.node_index].child[x][y][z] = link | 0x8000;
        ob->dirty_ = true;
      }
    }
    //  6. return success
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_octree_coord_get_box(ptf_geopos const & pos, ptf_16 level, 
    ptf_ocube * o_box)
{
  if (level > 30) {
    return peInvalidArguments;
  }

  coord3 c3;
  ptf_err ret = c3.set(pos);
  if (ret != peSuccess) {
    return ret;
  }

  o_box->x_min = c3.x & (0xffffffffUL << (31-level));
  o_box->y_min = c3.y & (0xffffffffUL << (31-level));
  o_box->z_min = c3.z & (0xffffffffUL << (31-level));
  o_box->x_max = o_box->x_min + (1UL << (31-level));
  o_box->y_max = o_box->y_min + (1UL << (31-level));
  o_box->z_max = o_box->z_min + (1UL << (31-level));

  assert(o_box->x_min <= c3.x);
  assert(o_box->x_max > c3.x);
  assert(o_box->y_min <= c3.y);
  assert(o_box->y_max > c3.y);
  assert(o_box->z_min <= c3.z);
  assert(o_box->z_max > c3.z);

  return peSuccess;
}


PTF_API ptf_err ptf_octree_coord_get_box_natural(ptf_geopos const &min, 
    ptf_geopos const &max, ptf_ocube * o_box)
{
  if (max.x < min.x || max.y < min.y || max.z < min.z) return peInvalidArguments;
  double sideLength = std::max(max.x - min.x, std::max(max.y - min.y, max.z - min.z));
  if (sideLength > PTF_MAX_GEOPOS) return peInvalidArguments;
  if (min.x + sideLength > PTF_MAX_GEOPOS
      || min.y + sideLength > PTF_MAX_GEOPOS
      || min.z + sideLength > PTF_MAX_GEOPOS) {
    return peInvalidArguments;
  }
  //  figure out level by inspecting length
  //  don't use log here, for precision reasons
  double curLength = PTF_MAX_GEOPOS - PTF_MIN_GEOPOS;
  ptf_16 level = 0;
  ptf_32 mask = 0x80000000;
  while (level < 30 && curLength * 0.25f >= sideLength) {
    ++level;
    curLength *= 0.5f;
    mask >>= 1;
  }
  ptf_err ret = ptf_octree_coord_get_box(min, level, o_box);
  if (ret != peSuccess) return ret;
  ptf_geopos lower, upper;
  ptf_octree_box_get_coords(*o_box, &lower, &upper);
  //  off-by-half
  if (lower.x + curLength <= max.x) {
    o_box->x_min += (mask >> 1);
    o_box->x_max += (mask >> 1);
  }
  if (upper.x - curLength > min.x) {
    o_box->x_min -= (mask >> 1);
    o_box->x_max -= (mask >> 1);
  }
  if (lower.y + curLength <= max.y) {
    o_box->y_min += (mask >> 1);
    o_box->y_max += (mask >> 1);
  }
  if (upper.y - curLength > min.y) {
    o_box->y_min -= (mask >> 1);
    o_box->y_max -= (mask >> 1);
  }
  if (lower.z + curLength <= max.z) {
    o_box->z_min += (mask >> 1);
    o_box->z_max += (mask >> 1);
  }
  if (upper.z - curLength > min.z) {
    o_box->z_min -= (mask >> 1);
    o_box->z_max -= (mask >> 1);
  }
#if !defined(NDEBUG)
  ptf_octree_box_get_coords(*o_box, &lower, &upper);
  assert(lower.x <= min.x && upper.x >= max.x);
  assert(lower.y <= min.y && upper.y >= max.y);
  assert(lower.z <= min.z && upper.z >= max.z);
#endif
  return peSuccess;
}

ptf_err ptf_octree_box_get_coords(ptf_ocube const & box, ptf_geopos * o_lower, 
    ptf_geopos * o_upper)
{
  double const d = (1.0 / GEOPOS_RATIO);

  o_lower->x = box.x_min * d + PTF_MIN_GEOPOS;
  o_lower->y = box.y_min * d + PTF_MIN_GEOPOS;
  o_lower->z = box.z_min * d + PTF_MIN_GEOPOS;
  o_upper->x = box.x_max * d + PTF_MIN_GEOPOS;
  o_upper->y = box.y_max * d + PTF_MIN_GEOPOS;
  o_upper->z = box.z_max * d + PTF_MIN_GEOPOS;

  return peSuccess;
}

PTF_API ptf_err ptf_octree_box_set_coords(ptf_geopos const & o_lower, ptf_geopos const & o_upper,
    ptf_ocube * box)
{
  box->x_min = ( o_lower.x - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;
  box->y_min = ( o_lower.y - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;
  box->z_min = ( o_lower.z - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;
  box->x_max = ( o_upper.x - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;
  box->y_max = ( o_upper.y - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;
  box->z_max = ( o_upper.z - PTF_MIN_GEOPOS ) * GEOPOS_RATIO;

  return peSuccess;
}

ptf_err ptf_meters_to_octreecoord( double meters, ptf_32 * geo_units )
{
  try {
    *geo_units = ( meters - (double) PTF_MIN_GEOPOS ) * (double) GEOPOS_RATIO;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_octreecoord_to_meters( ptf_32 geo_units, double * meters )
{
  try {
    double const d = (1.0 / GEOPOS_RATIO);
    *meters = geo_units * d + PTF_MIN_GEOPOS;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_level_to_box_side( ptf_32 level, ptf_32 * side )
{
  if (level > 30) {
    return peInvalidArguments;
  }

  *side = (1UL << (31-level));

  return peSuccess;
}

ptf_err ptf_box_side_to_level( ptf_32 side , ptf_32 * level )
{
  assert(!(side & (side - 1)));
  ptf_32 mask = 1;
  ptf_16 lvl = 31;
  while (!(mask & side) && lvl > 0) {
    mask <<= 1;
    lvl--;
  }
  *level = lvl;
  return peSuccess;
}

bool ptf_ocube_equals_ocube(ptf_ocube const &a, ptf_ocube const &b)
{
  return a.x_min == b.x_min
      && a.x_max == b.x_max
      && a.y_min == b.y_min
      && a.y_max == b.y_max
      && a.z_min == b.z_min
      && a.z_max == b.z_max;
}

//  a == big
//  b == small
bool ptf_ocube_contains_ocube(ptf_ocube const &a, ptf_ocube const &b)
{
  return a.x_min <= b.x_min
      && a.x_max >= b.x_max
      && a.y_min <= b.y_min
      && a.y_max >= b.y_max
      && a.z_min <= b.z_min
      && a.z_max >= b.z_max;
}

