#include "ptf_spheroid.h"

#include <string.h>
#include <assert.h>

class ptf_spheroid_t {
  public:
    ptf_spheroid_data header_;
    ptf_32            sig_;
    
    ptf_spheroid_t() : sig_('sph')
    {
      memset( &header_, 0, sizeof( header_ ) );
    }

    ~ptf_spheroid_t() { clear(); }
    bool validate()   { return sig_ == 'sph'; }
    void clear()
    {
      sig_     = 0;
      memset( &header_, 0, sizeof( header_ ) );
    }
};


static ptf_err validate_header(ptf_spheroid_data const * header)
{
  if( header->material_id >= 0x0100 ) {
    return peTooBigForMachine;
  }
  return peSuccess;
}

ptf_err ptf_spheroid_create(ptf_spheroid_handle *o_spheroid, ptf_char const *name)
{
  try {
    size_t len = strlen((char const *)name);
    if (len >= sizeof((*o_spheroid)->header_.name)) {
      return peInvalidArguments;
    }
    *o_spheroid = new ptf_spheroid_t();
    memcpy((*o_spheroid)->header_.name, name, len+1);
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_spheroid_destroy(ptf_spheroid_handle spheroid)
{
  try {
    if (!spheroid->validate()) 
      return peInvalidArguments;

    delete spheroid;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_spheroid_read_from_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    if( !spheroid->validate() )
      return peInvalidArguments;

    ptf_err ptf_status;
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptSpheroidData;
    ref.index = fref.index;

    // read the header from the file
    ptf_status = ptf_file_read_record( file, &ref, 0, &spheroid->header_, sizeof( spheroid->header_ ) );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }
    ptf_status = validate_header( &spheroid->header_ );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_spheroid_update_to_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    // set the "current record" in the PTF file
    ptf_err ptf_status;
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptSpheroidData;
    ref.index = fref.index;

    ptf_status = ptf_file_replace_record( file, &ref );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }

    // write out the header
    ptf_status = ptf_file_write_record( file, &spheroid->header_, sizeof( spheroid->header_ ) );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }

    // commit record to disk
    ptf_record_ref actualRef;
    memset( &actualRef, 0, sizeof( actualRef ) );
    ptf_status = ptf_file_commit_record( file, &actualRef );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }

#ifdef _DEBUG
    // verify that the record ref we get back from commit has the expected index
    if ( !( actualRef.type == ptSpheroidData && actualRef.index == ref.index ) ) {
      return peIoError;
    }
#endif

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_spheroid_append_to_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref *o_ref)
{
  try {
    // add a new raster record to the PTF file
    ptf_err ptf_status;
    ptf_status = ptf_file_new_record( file, ptSpheroidData );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }

    // write out the header
    ptf_status = ptf_file_write_record( file, &spheroid->header_, sizeof( spheroid->header_ ) );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }

    // commit record to disk
    ptf_record_ref actualRef;
    memset( &actualRef, 0, sizeof( actualRef ) );
    ptf_status = ptf_file_commit_record( file, &actualRef );
    if( ptf_status != peSuccess ) {
      return ptf_status;
    }
    if ( o_ref ) {
      o_ref->index = actualRef.index;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_spheroid_set_data(ptf_spheroid_handle spheroid, ptf_spheroid_data const &data)
{
  try {
    if( !spheroid->validate() ) {
      return peInvalidArguments;
    }
    memcpy( &spheroid->header_, &data, sizeof( ptf_spheroid_data ) );
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_spheroid_get_data(ptf_spheroid_handle spheroid, ptf_spheroid_data * o_data)
{
  try {
    if( !spheroid->validate() ) {
      return peInvalidArguments;
    }
    memcpy( o_data, &spheroid->header_, sizeof( ptf_spheroid_data ) );
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}
