
#include "ptf_str.h"
#include <ctype.h>
#include <algorithm>

#undef min
#undef max


static size_t one_utf8(wchar_t uc, ptf_char * & dst, size_t & count_dst)
{
  size_t n = 1, r;
  if (uc >= (1U<<7)) n++;
  if (uc >= (1U<<12)) n++;
  if (uc >= (1U<<17)) n++;
  if (uc >= (1U<<22)) n++;
  if (uc >= (1U<<29)) n++;
  r = n;
  // test with "<" to make sure there's always space to terminate
  if (dst && n < count_dst) {
    *dst++ = (ptf_char)(((0xfeU) << (8-n)) | (uc >> (n-1)*6));
    while (n > 1) {
      --n;
      *dst++ = (ptf_char)((0x80) | ((uc >> (n-1)*6) & 0x3f));
    }
    count_dst -= r;
  }
  else if (dst && count_dst > 0) {
    *dst = 0;
    dst += count_dst;
    count_dst = 0;
  }
  return r;
}

template<typename T, size_t M>
bool one(ptf_char const * & src, T * & dst, size_t & dst_count, T stand_in)
{
  unsigned char ch = *src;
  if (!ch) {
    return false;
  }
  size_t n = 1;
  if ((ch & 0xE0) == 0xC0) n = 2;
  else if ((ch & 0xF0) == 0xE0) n = 3;
  else if ((ch & 0xF8) == 0xF0) n = 4;
  else if ((ch & 0xFC) == 0xF8) n = 5;
  else if ((ch & 0xFE) == 0xFC) n = 6;
  else if ((ch & 0x80) == 0x80) {
    return false;
  }
  ++src; --n;
  ptf_32 s = ch & (0x7f >> n);
  while (n > 0) {
    ch = *src;
    if ((ch & 0xC0) != 0x80) {
      return false;
    }
    ++src; --n;
    s = (s<<6) | (ch & 0x3F);
  }
  if (s > M) {
    *dst = stand_in;
  }
  else {
    *dst = (T)s;
  }
  if (dst && dst_count > 1) {
    ++dst;
    --dst_count;
  }
  return true;
}

size_t ptf_str_to_char(ptf_char const * src, char * dst, size_t count_dst, char stand_in)
{
  size_t u = 0;
  while (*src) {
    if (!one<char,255>(src, dst, count_dst, stand_in)) {
      goto invalid;
    }
    ++u;
  }
  if (dst && count_dst) {
    *dst = 0;
  }
  return u+1;
invalid:
  return 0;
}

size_t ptf_str_to_wchar(ptf_char const * src, wchar_t * dst, size_t count_dst, wchar_t stand_in)
{
  size_t u = 0;
  while (*src) {
    if (!one<wchar_t,0xffff>(src, dst, count_dst, stand_in)) {
      goto invalid;
    }
    ++u;
  }
  if (dst && count_dst) {
    *dst = 0;
  }
  return u+1;
invalid:
  if (dst && count_dst) *dst = 0;
  return 0;
}

size_t ptf_str_from_char(char const * str, ptf_char * dst, size_t count_dst)
{
  size_t u = 0;
  while (*str) {
    unsigned char uc = *(unsigned char *)str;
    ++str;
    u += one_utf8(uc, dst, count_dst);
  }
  if (dst && count_dst) {
    *dst = 0;
  }
  return u+1;
}

size_t ptf_str_from_wchar(wchar_t const * str, ptf_char * dst, size_t count_dst)
{
  size_t u = 0;
  while (*str) {
    wchar_t uc = *str;
    ++str;
    u += one_utf8(uc, dst, count_dst);
  }
  if (dst && count_dst) {
    *dst = 0;
  }
  return u+1;
}

size_t ptf_str_basename(ptf_char const * path, ptf_char * dst, size_t size)
{
  ptf_char const * last = path-1;
  while (*path) {
    if (*path == '/' || *path == '\\') {
      last = path;
    }
    ++path;
  }
  if (dst) {
    if ((size_t)(path-last) >= size) {
      if (size > 1) {
        memcpy(dst, last+1, size-1);
      }
      if (size) {
        dst[size-1] = 0;
      }
    }
    else {
      memcpy(dst, last+1, path-last);
    }
  }
  return path-last;
}

//  TODO: this function should be improved to do proper 
//  locale-sensitive collation. However, that becomes a 
//  major portability hassle.
int ptf_strcasecmp(ptf_char const * a, ptf_char const * b)
{
  while (*a && *b) {
    int ca = isupper(*a) ? tolower(*a) : *a;
    int cb = isupper(*b) ? tolower(*b) : *b;
    if (ca < cb) return -1;
    if (ca > cb) return 1;
    ++a;
    ++b;
  }
  if (*a) return 1;
  if (*b) return -1;
  return 0;
}

int ptf_strncpy(ptf_char * dst, size_t size, ptf_char const * src)
{
  if (size == 0) {
    //  can't zero-terminate when there is no space
    return -1;
  }
  ptf_char const * bottom = src;
  while (size > 1 && *src) {
    *dst++ = *src++;
    --size;
  }
  if (!*src) {
    *dst = 0;
    return 0;       //  proper copy made
  }
  if (*src < 128 || *src >= 192) { //  stopped on a glyph boundary
    *dst = 0;
    return -1;
  }
  //  For properly formatted UTF-8, bytes in the range [128,192) 
  //  are always preceded by another such byte, back to the start 
  //  byte, which is guaranteed to be in [192,255).
  //  For improperly formatted UTF-8, do not under-run the buffer.
  while (src > bottom && *src < 192) {  //  find start of glyph
    --src;
    --dst;
  }
  *dst = 0;
  return -1;
}

